from __future__ import annotations

from typing import Type, Optional, Any, TypeVar, Callable, overload, Iterator, Tuple, ClassVar, Dict

__all__ = (
    'Intents',
)


FV = TypeVar('FV', bound='flag_value')
BF = TypeVar('BF', bound='BaseFlags')


class flag_value:
    def __init__(self, func: Callable[[Any], int]):
        self.flag = func(None)
        self.__doc__ = func.__doc__

    @overload
    def __get__(self: FV, instance: None, owner: Type[BF]) -> FV:
        ...

    @overload
    def __get__(self, instance: BF, owner: Type[BF]) -> bool:
        ...

    def __get__(self, instance: Optional[BF], owner: Type[BF]) -> Any:
        if instance is None:
            return self
        return instance._has_flag(self.flag)

    def __set__(self, instance: BF, value: bool) -> None:
        instance._set_flag(self.flag, value)

    def __repr__(self):
        return f'<flag_value flag={self.flag!r}>'


class alias_flag_value(flag_value):
    pass


def fill_with_flags(*, inverted: bool = False):
    def decorator(cls: Type[BF]):
        # fmt: off
        cls.VALID_FLAGS = {
            name: value.flag
            for name, value in cls.__dict__.items()
            if isinstance(value, flag_value)
        }
        # fmt: on

        if inverted:
            max_bits = max(cls.VALID_FLAGS.values()).bit_length()
            cls.DEFAULT_VALUE = -1 + (2 ** max_bits)
        else:
            cls.DEFAULT_VALUE = 0

        return cls

    return decorator


# n.b. flags must inherit from this and use the decorator above
class BaseFlags:
    VALID_FLAGS: ClassVar[Dict[str, int]]
    DEFAULT_VALUE: ClassVar[int]

    value: int

    __slots__ = ('value',)

    def __init__(self, **kwargs: bool):
        self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f'{key!r} is not a valid flag name.')
            setattr(self, key, value)

    @classmethod
    def _from_value(cls, value):
        self = cls.__new__(cls)
        self.value = value
        return self

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} value={self.value}>'

    def __iter__(self) -> Iterator[Tuple[str, bool]]:
        for name, value in self.__class__.__dict__.items():
            if isinstance(value, alias_flag_value):
                continue

            if isinstance(value, flag_value):
                yield name, self._has_flag(value.flag)

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) == o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value |= o
        elif toggle is False:
            self.value &= ~o
        else:
            raise TypeError(f'Value to set for {self.__class__.__name__} must be a bool.')


@fill_with_flags()
class Intents(BaseFlags):
    __slots__ = ()

    def __init__(self, **kwargs: bool):
        self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f'{key!r} is not a valid flag name.')
            setattr(self, key, value)

    @classmethod
    def all(cls: Type[Intents]) -> Intents:
        """一个工厂方法，它创建一个 :class:`Intents` 并启用所有内容。"""
        value = 0
        for bits in cls.VALID_FLAGS.values():
            value |= bits
        self = cls.__new__(cls)
        self.value = value
        return self

    @classmethod
    def none(cls: Type[Intents]) -> Intents:
        """一个工厂方法，它创建一个禁用一切的 :class:`Intents` 。"""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @classmethod
    def default(cls: Type[Intents]) -> Intents:
        """一个工厂方法，它创建一个 :class:`Intents`，
        除了 :attr:`guilds`, :attr:`members` 和 :attr:`at_guild_messages` 之外的所有内容都禁用。
        """
        self = cls.none()
        self.guilds = True
        self.members = True
        self.at_guild_messages = True
        return self

    @flag_value
    def guilds(self):
        """:class:`bool`: 频道相关事件是否开启。

        这对应于以下事件：

        - :func:`on_guild_join`
        - :func:`on_guild_update`
        - :func:`on_guild_remove`
        - :func:`on_guild_channel_update`
        - :func:`on_guild_channel_create`
        - :func:`on_guild_channel_delete`

        这也对应于缓存方面的以下属性和类：

        - :attr:`Client.guilds`
        - :class:`Guild` 以及它的所有属性。
        - :meth:`Client.get_channel`
        - :meth:`Client.get_all_channels`

        .. warning::

            强烈建议您启用此意图，以便您的机器人正常运行。
        """

        return 1 << 0

    @flag_value
    def members(self):
        """:class:`bool`: 频道成员相关事件是否开启。

        这对应于以下事件：

        - :func:`on_member_join`
        - :func:`on_member_remove`
        - :func:`on_member_update`

        这也对应于缓存方面的以下属性和类：

        - :meth:`Client.get_all_members`
        - :meth:`Client.get_user`
        - :meth:`Guild.chunk`
        - :meth:`Guild.fetch_members`
        - :meth:`Guild.get_member`
        - :attr:`Guild.members`
        - :attr:`Member.roles`
        - :attr:`Member.nick`
        - :attr:`User.name`
        - :attr:`User.avatar`
        """
        return 1 << 1

    @alias_flag_value
    def messages(self):
        """:class:`bool`: 是否启用频道和直接消息相关事件。
        这是设置或获取 :attr:`guild_messages` 和 :attr:`dm_messages` 的快捷方式。

        这对应于以下事件：
        - :func:`on_message` (both guilds and DMs)

        这也对应于缓存方面的以下属性和类：
        - :class:`Message`
        - :attr:`Client.cached_messages`

        .. note::

            现在来说，这个 Intents 需要额外的申请，如果没有适当的权限 Websocket 将无法连接。
        """

        return (1 << 12) | (1 << 9)

    @flag_value
    def dm_messages(self):
        """:class:`bool`: 是否启用直接消息相关事件。
        另见 :attr:`guild_messages` 来获取频道信息或 :attr:`messages` 同时获取两者。

        这对应于以下事件：

        - :func:`on_message` (only for DMs)

        这也对应于缓存方面的以下属性和类：

        - :class:`Message`
        - :attr:`Client.cached_messages` (only for DMs)

        .. note::

            现在来说，这个 Intents 需要额外的申请，如果没有适当的权限 Websocket 将无法连接。
        """
        return 1 << 12

    @flag_value
    def audio(self):
        """:class:`bool`: 频道提及机器人的消息相关事件是否开启。

        这对应于以下事件：

        - :func:`on_audio_start`
        - :func:`on_audio_stop`
        - :func:`on_mic_start`
        - :func:`on_mic_stop`
        """
        return 1 << 29

    @flag_value
    def at_guild_messages(self):
        """:class:`bool`: 频道提及机器人的消息相关事件是否开启。

        这对应于以下事件：

        - :func:`on_message` (只适用于频道)

        这也对应于缓存方面的以下属性和类：

        - :class:`Message`
        - :attr:`Client.cached_messages` (只适用于频道)
        """
        return 1 << 30

    @flag_value
    def guild_messages(self):
        """:class:`bool`: 频道消息相关事件是否开启。

        这对应于以下事件：

        - :func:`on_message` (只适用于频道)

        这也对应于缓存方面的以下属性和类：

        - :class:`Message`
        - :attr:`Client.cached_messages` (只适用于频道)

        .. note::

            现在来说，这个 Intents 需要额外的申请，如果没有适当的权限 Websocket 将无法连接。
        """
        return 1 << 9

    @flag_value
    def guild_reactions(self):
        """:class:`bool`: 频道消息反应相关事件是否开启。

        这对应于以下事件：

        - :func:`on_reaction_add` (只适用于频道)
        - :func:`on_reaction_remove` (只适用于频道)
        - :func:`on_raw_reaction_add` (只适用于频道)
        - :func:`on_raw_reaction_remove` (只适用于频道)

        这也对应于缓存方面的以下属性和类：

        - :attr:`Message.reactions` (只适用于频道信息)

        .. note::

            现在来说，这个 Intents 需要额外的申请，如果没有适当的权限 Websocket 将无法连接。
        """
        return 1 << 10

    @flag_value
    def threads(self):
        """:class:`bool`: 论坛频道相关事件是否开启。

        .. note::

            现在来说，这个 Intents 需要额外的申请，如果没有适当的权限 Websocket 将无法连接。
        """
        return 1 << 28



from __future__ import annotations

from . import utils
from .mixins import Hashable

from typing import (
    SupportsInt,
    TYPE_CHECKING,
    Union,
)

if TYPE_CHECKING:
    SupportsIntCast = Union[SupportsInt, str, bytes, bytearray]

__all__ = (
    'Object',
)


class Object(Hashable):
    """代表一个通用的 QQ 对象。
    这个类的目的是允许您创建数据类的 ``微型`` 版，让你只传入一个 ID。
    大多数接受带有 ID 的特定数据类的函数也可以接受这个类作为替代。
    请注意，即使是这种情况，并非所有对象（如果有）实际上都继承自此类。
    在某些情况下，某些 websocket 事件以奇怪的顺序接收，当发生此类事件时，
    您将收到此类而不是实际的数据类。这些情况极为罕见。

    .. container:: operations

        .. describe:: x == y

            检查两个对象是否相等。

        .. describe:: x != y

            检查两个对象是否不相等。

        .. describe:: hash(x)

            返回对象的哈希值。

    Attributes
    -----------
    id: :class:`int`
        对象的 ID。
    """
    def __init__(self, id: SupportsIntCast):
        try:
            id = int(id)
        except ValueError:
            raise TypeError(f'id 参数必须可转换为 int 而不是 {id.__class__!r}') from None
        else:
            self.id = id

    def __repr__(self) -> str:
        return f'<Object id={self.id!r}>'



from __future__ import annotations

import time
import random
from typing import Callable, Generic, Literal, TypeVar, overload, Union

T = TypeVar('T', bool, Literal[True], Literal[False])

__all__ = (
    'ExponentialBackoff',
)


class ExponentialBackoff(Generic[T]):
    def __init__(self, base: int = 1, *, integral: T = False):
        self._base: int = base

        self._exp: int = 0
        self._max: int = 10
        self._reset_time: int = base * 2 ** 11
        self._last_invocation: float = time.monotonic()

        # Use our own random instance to avoid messing with global one
        rand = random.Random()
        rand.seed()

        self._randfunc: Callable[..., Union[int, float]] = rand.randrange if integral else rand.uniform  # type: ignore

    @overload
    def delay(self: ExponentialBackoff[Literal[False]]) -> float:
        ...

    @overload
    def delay(self: ExponentialBackoff[Literal[True]]) -> int:
        ...

    @overload
    def delay(self: ExponentialBackoff[bool]) -> Union[int, float]:
        ...

    def delay(self) -> Union[int, float]:
        """计算下一个延迟
        根据指数退避算法返回要等待的下一个延迟。
        这是一个介于 0 和 基数 * 2^exp 之间的值，其中 exp 从 1 开始，并在每次调用此方法时递增，最大为 10。
        如果自上次重试以来已经过去了超过基数 * 2^11 的时间段，则指数将重置为 1。
        """
        invocation = time.monotonic()
        interval = invocation - self._last_invocation
        self._last_invocation = invocation

        if interval > self._reset_time:
            self._exp = 0

        self._exp = min(self._exp + 1, self._max)
        return self._randfunc(0, self._base * 2 ** self._exp)



from __future__ import annotations

from .errors import (
    BadFlagArgument,
    CommandError,
    MissingFlagArgument,
    TooManyFlags,
    MissingRequiredFlag,
)

from qq.utils import resolve_annotation
from .view import StringView
from .converter import run_converters

from qq.utils import maybe_coroutine, MISSING
from dataclasses import dataclass, field
from typing import (
    Dict,
    Iterator,
    Literal,
    Optional,
    Pattern,
    Set,
    TYPE_CHECKING,
    Tuple,
    List,
    Any,
    Type,
    TypeVar,
    Union,
)

import inspect
import sys
import re

__all__ = (
    'Flag',
    'flag',
    'FlagConverter',
)

if TYPE_CHECKING:
    from .context import Context


@dataclass
class Flag:
    """表示 FlagConverter 的标志参数。

    :func:`~qq.ext.commands.flag` 函数有助于创建这些标志对象，但没有必要这样做。这些不能手动构建。

    Attributes
    ------------
    name: :class:`str`
        标志的名称。
    aliases: List[:class:`str`]
        标志名称的别名。
    attribute: :class:`str`
        与此标志对应的类中的属性。
    default: Any
        标志的默认值（如果可用）。
    annotation: Any
        标志的基础评估注释。
    max_args: :class:`int`
        标志可以接受的最大参数数。负值表示无限数量的参数。
    override: :class:`bool`
        多个给定值是否覆盖先前的值。
    """

    name: str = MISSING
    aliases: List[str] = field(default_factory=list)
    attribute: str = MISSING
    annotation: Any = MISSING
    default: Any = MISSING
    max_args: int = MISSING
    override: bool = MISSING
    cast_to_dict: bool = False

    @property
    def required(self) -> bool:
        """:class:`bool`: 是否需要标志。

        必需标志没有默认值。
        """
        return self.default is MISSING


def flag(
        *,
        name: str = MISSING,
        aliases: List[str] = MISSING,
        default: Any = MISSING,
        max_args: int = MISSING,
        override: bool = MISSING,
) -> Any:
    """覆盖底层 FlagConverter 类属性的默认功能和参数。

    Parameters
    ------------
    name: :class:`str`
        标志名。如果未给出，则默认为属性名称。
    aliases: List[:class:`str`]
        标志名称的别名。如果没有给出，则不会设置别名。
    default: Any
        默认参数。这可以是一个值，也可以是一个以 :class:`Context` 作为其唯一参数的可调用对象。如果未给出，则默认为赋予属性的默认值。
    max_args: :class:`int`
        标志可以接受的最大参数数。负值表示无限数量的参数。默认值取决于给定的注释。
    override: :class:`bool`
        多个给定值是否覆盖先前的值。默认值取决于给定的注释。
    """
    return Flag(name=name, aliases=aliases, default=default, max_args=max_args, override=override)


def validate_flag_name(name: str, forbidden: Set[str]):
    if not name:
        raise ValueError('标志名称不应为空')

    for ch in name:
        if ch.isspace():
            raise ValueError(f'标志名称 {name!r} 不能有空格')
        if ch == '\\':
            raise ValueError(f'标志名称 {name!r} 不能有反斜杠')
        if ch in forbidden:
            raise ValueError(f'标志名称 {name!r} 不能包含任何 {forbidden!r}')


def get_flags(namespace: Dict[str, Any], globals: Dict[str, Any], locals: Dict[str, Any]) -> Dict[str, Flag]:
    annotations = namespace.get('__annotations__', {})
    case_insensitive = namespace['__commands_flag_case_insensitive__']
    flags: Dict[str, Flag] = {}
    cache: Dict[str, Any] = {}
    names: Set[str] = set()
    for name, annotation in annotations.items():
        flag = namespace.pop(name, MISSING)
        if isinstance(flag, Flag):
            flag.annotation = annotation
        else:
            flag = Flag(name=name, annotation=annotation, default=flag)

        flag.attribute = name
        if flag.name is MISSING:
            flag.name = name

        annotation = flag.annotation = resolve_annotation(flag.annotation, globals, locals, cache)

        if flag.default is MISSING and hasattr(annotation,
                                               '__commands_is_flag__') and annotation._can_be_constructible():
            flag.default = annotation._construct_default

        if flag.aliases is MISSING:
            flag.aliases = []

        # Add sensible defaults based off of the type annotation
        # <type> -> (max_args=1)
        # List[str] -> (max_args=-1)
        # Tuple[int, ...] -> (max_args=1)
        # Dict[K, V] -> (max_args=-1, override=True)
        # Union[str, int] -> (max_args=1)
        # Optional[str] -> (default=None, max_args=1)

        try:
            origin = annotation.__origin__
        except AttributeError:
            # A regular type hint
            if flag.max_args is MISSING:
                flag.max_args = 1
        else:
            if origin is Union:
                # typing.Union
                if flag.max_args is MISSING:
                    flag.max_args = 1
                if annotation.__args__[-1] is type(None) and flag.default is MISSING:
                    # typing.Optional
                    flag.default = None
            elif origin is tuple:
                # typing.Tuple
                # tuple parsing is e.g. `flag: peter 20`
                # for Tuple[str, int] would give you flag: ('peter', 20)
                if flag.max_args is MISSING:
                    flag.max_args = 1
            elif origin is list:
                # typing.List
                if flag.max_args is MISSING:
                    flag.max_args = -1
            elif origin is dict:
                # typing.Dict[K, V]
                # Equivalent to:
                # typing.List[typing.Tuple[K, V]]
                flag.cast_to_dict = True
                if flag.max_args is MISSING:
                    flag.max_args = -1
                if flag.override is MISSING:
                    flag.override = True
            elif origin is Literal:
                if flag.max_args is MISSING:
                    flag.max_args = 1
            else:
                raise TypeError(f'{flag.name!r} 标志不支持键入注释 {annotation!r}')

        if flag.override is MISSING:
            flag.override = False

        # Validate flag names are unique
        name = flag.name.casefold() if case_insensitive else flag.name
        if name in names:
            raise TypeError(f'{flag.name!r} 标志与之前的标志或别名冲突。')
        else:
            names.add(name)

        for alias in flag.aliases:
            # Validate alias is unique
            alias = alias.casefold() if case_insensitive else alias
            if alias in names:
                raise TypeError(f'{flag.name!r} 标志别名 {alias!r} 与之前的标志或别名冲突。')
            else:
                names.add(alias)

        flags[flag.name] = flag

    return flags


class FlagsMeta(type):
    if TYPE_CHECKING:
        __commands_is_flag__: bool
        __commands_flags__: Dict[str, Flag]
        __commands_flag_aliases__: Dict[str, str]
        __commands_flag_regex__: Pattern[str]
        __commands_flag_case_insensitive__: bool
        __commands_flag_delimiter__: str
        __commands_flag_prefix__: str

    def __new__(
            cls: Type[type],
            name: str,
            bases: Tuple[type, ...],
            attrs: Dict[str, Any],
            *,
            case_insensitive: bool = MISSING,
            delimiter: str = MISSING,
            prefix: str = MISSING,
    ):
        attrs['__commands_is_flag__'] = True

        try:
            global_ns = sys.modules[attrs['__module__']].__dict__
        except KeyError:
            global_ns = {}

        frame = inspect.currentframe()
        try:
            if frame is None:
                local_ns = {}
            else:
                if frame.f_back is None:
                    local_ns = frame.f_locals
                else:
                    local_ns = frame.f_back.f_locals
        finally:
            del frame

        flags: Dict[str, Flag] = {}
        aliases: Dict[str, str] = {}
        for base in reversed(bases):
            if base.__dict__.get('__commands_is_flag__', False):
                flags.update(base.__dict__['__commands_flags__'])
                aliases.update(base.__dict__['__commands_flag_aliases__'])
                if case_insensitive is MISSING:
                    attrs['__commands_flag_case_insensitive__'] = base.__dict__['__commands_flag_case_insensitive__']
                if delimiter is MISSING:
                    attrs['__commands_flag_delimiter__'] = base.__dict__['__commands_flag_delimiter__']
                if prefix is MISSING:
                    attrs['__commands_flag_prefix__'] = base.__dict__['__commands_flag_prefix__']

        if case_insensitive is not MISSING:
            attrs['__commands_flag_case_insensitive__'] = case_insensitive
        if delimiter is not MISSING:
            attrs['__commands_flag_delimiter__'] = delimiter
        if prefix is not MISSING:
            attrs['__commands_flag_prefix__'] = prefix

        case_insensitive = attrs.setdefault('__commands_flag_case_insensitive__', False)
        delimiter = attrs.setdefault('__commands_flag_delimiter__', ':')
        prefix = attrs.setdefault('__commands_flag_prefix__', '')

        for flag_name, flag in get_flags(attrs, global_ns, local_ns).items():
            flags[flag_name] = flag
            aliases.update({alias_name: flag_name for alias_name in flag.aliases})

        forbidden = set(delimiter).union(prefix)
        for flag_name in flags:
            validate_flag_name(flag_name, forbidden)
        for alias_name in aliases:
            validate_flag_name(alias_name, forbidden)

        regex_flags = 0
        if case_insensitive:
            flags = {key.casefold(): value for key, value in flags.items()}
            aliases = {key.casefold(): value.casefold() for key, value in aliases.items()}
            regex_flags = re.IGNORECASE

        keys = list(re.escape(k) for k in flags)
        keys.extend(re.escape(a) for a in aliases)
        keys = sorted(keys, key=lambda t: len(t), reverse=True)

        joined = '|'.join(keys)
        pattern = re.compile(f'(({re.escape(prefix)})(?P<flag>{joined}){re.escape(delimiter)})', regex_flags)
        attrs['__commands_flag_regex__'] = pattern
        attrs['__commands_flags__'] = flags
        attrs['__commands_flag_aliases__'] = aliases

        return type.__new__(cls, name, bases, attrs)


async def tuple_convert_all(ctx: Context, argument: str, flag: Flag, converter: Any) -> Tuple[Any, ...]:
    view = StringView(argument)
    results = []
    param: inspect.Parameter = ctx.current_parameter  # type: ignore
    while not view.eof:
        view.skip_ws()
        if view.eof:
            break

        word = view.get_quoted_word()
        if word is None:
            break

        try:
            converted = await run_converters(ctx, converter, word, param)
        except CommandError:
            raise
        except Exception as e:
            raise BadFlagArgument(flag) from e
        else:
            results.append(converted)

    return tuple(results)


async def tuple_convert_flag(ctx: Context, argument: str, flag: Flag, converters: Any) -> Tuple[Any, ...]:
    view = StringView(argument)
    results = []
    param: inspect.Parameter = ctx.current_parameter  # type: ignore
    for converter in converters:
        view.skip_ws()
        if view.eof:
            break

        word = view.get_quoted_word()
        if word is None:
            break

        try:
            converted = await run_converters(ctx, converter, word, param)
        except CommandError:
            raise
        except Exception as e:
            raise BadFlagArgument(flag) from e
        else:
            results.append(converted)

    if len(results) != len(converters):
        raise BadFlagArgument(flag)

    return tuple(results)


async def convert_flag(ctx, argument: str, flag: Flag, annotation: Any = None) -> Any:
    param: inspect.Parameter = ctx.current_parameter  # type: ignore
    annotation = annotation or flag.annotation
    try:
        origin = annotation.__origin__
    except AttributeError:
        pass
    else:
        if origin is tuple:
            if annotation.__args__[-1] is Ellipsis:
                return await tuple_convert_all(ctx, argument, flag, annotation.__args__[0])
            else:
                return await tuple_convert_flag(ctx, argument, flag, annotation.__args__)
        elif origin is list:
            # typing.List[x]
            annotation = annotation.__args__[0]
            return await convert_flag(ctx, argument, flag, annotation)
        elif origin is Union and annotation.__args__[-1] is type(None):
            # typing.Optional[x]
            annotation = Union[annotation.__args__[:-1]]
            return await run_converters(ctx, annotation, argument, param)
        elif origin is dict:
            # typing.Dict[K, V] -> typing.Tuple[K, V]
            return await tuple_convert_flag(ctx, argument, flag, annotation.__args__)

    try:
        return await run_converters(ctx, annotation, argument, param)
    except CommandError:
        raise
    except Exception as e:
        raise BadFlagArgument(flag) from e


F = TypeVar('F', bound='FlagConverter')


class FlagConverter(metaclass=FlagsMeta):
    """允许用户友好标志语法的转换器。

    这些标志是使用 :pep:`526` 类型注释定义的，类似于 Python 模块:mod:`dataclasses`。
    有关此转换器如何工作的更多信息，请查看相应的 :ref:`文档 <ext_commands_flag_converter>` 。

    .. container:: operations

        .. describe:: iter(x)

            返回 ``(flag_name, flag_value)`` 对的迭代器。
            例如，这允许将其构造为字典或对列表。请注意，未显示别名。

    Parameters
    -----------
    case_insensitive: :class:`bool`
        用于切换标志解析不区分大小写的类参数。如果为 ``True`` ，则以不区分大小写的方式解析标志。默认为 ``False`` 。
    prefix: :class:`str`
        所有标志都必须带有前缀的前缀。默认情况下没有前缀。
    delimiter: :class:`str`
        将标志的参数与标志的名称分开的分隔符。默认情况下，这是 ``:`` 。
    """

    @classmethod
    def get_flags(cls) -> Dict[str, Flag]:
        """Dict[:class:`str`, :class:`Flag`]: 标志名称到此转换器具有的标志对象的映射。"""
        return cls.__commands_flags__.copy()

    @classmethod
    def _can_be_constructible(cls) -> bool:
        return all(not flag.required for flag in cls.__commands_flags__.values())

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        for flag in self.__class__.__commands_flags__.values():
            yield (flag.name, getattr(self, flag.attribute))

    @classmethod
    async def _construct_default(cls: Type[F], ctx: Context) -> F:
        self: F = cls.__new__(cls)
        flags = cls.__commands_flags__
        for flag in flags.values():
            if callable(flag.default):
                default = await maybe_coroutine(flag.default, ctx)
                setattr(self, flag.attribute, default)
            else:
                setattr(self, flag.attribute, flag.default)
        return self

    def __repr__(self) -> str:
        pairs = ' '.join([f'{flag.attribute}={getattr(self, flag.attribute)!r}' for flag in self.get_flags().values()])
        return f'<{self.__class__.__name__} {pairs}>'

    @classmethod
    def parse_flags(cls, argument: str) -> Dict[str, List[str]]:
        result: Dict[str, List[str]] = {}
        flags = cls.__commands_flags__
        aliases = cls.__commands_flag_aliases__
        last_position = 0
        last_flag: Optional[Flag] = None

        case_insensitive = cls.__commands_flag_case_insensitive__
        for match in cls.__commands_flag_regex__.finditer(argument):
            begin, end = match.span(0)
            key = match.group('flag')
            if case_insensitive:
                key = key.casefold()

            if key in aliases:
                key = aliases[key]

            flag = flags.get(key)
            if last_position and last_flag is not None:
                value = argument[last_position: begin - 1].lstrip()
                if not value:
                    raise MissingFlagArgument(last_flag)

                try:
                    values = result[last_flag.name]
                except KeyError:
                    result[last_flag.name] = [value]
                else:
                    values.append(value)

            last_position = end
            last_flag = flag

        # Add the remaining string to the last available flag
        if last_position and last_flag is not None:
            value = argument[last_position:].strip()
            if not value:
                raise MissingFlagArgument(last_flag)

            try:
                values = result[last_flag.name]
            except KeyError:
                result[last_flag.name] = [value]
            else:
                values.append(value)

        # Verification of values will come at a later stage
        return result

    @classmethod
    async def convert(cls: Type[F], ctx: Context, argument: str) -> F:
        """|coro|

        实际将参数转换为标志映射的方法。

        Parameters
        ----------
        cls: Type[:class:`FlagConverter`]
            标志转换器类。
        ctx: :class:`Context`
            调用 context 。
        argument: :class:`str`
            要转换的参数。

        Raises
        --------
        FlagError
            与标志相关的解析错误。
        CommandError
            命令相关的错误。

        Returns
        --------
        :class:`FlagConverter`
            已解析所有标志的标志转换器实例。
        """
        arguments = cls.parse_flags(argument)
        flags = cls.__commands_flags__

        self: F = cls.__new__(cls)
        for name, flag in flags.items():
            try:
                values = arguments[name]
            except KeyError:
                if flag.required:
                    raise MissingRequiredFlag(flag)
                else:
                    if callable(flag.default):
                        default = await maybe_coroutine(flag.default, ctx)
                        setattr(self, flag.attribute, default)
                    else:
                        setattr(self, flag.attribute, flag.default)
                    continue

            if flag.max_args > 0 and len(values) > flag.max_args:
                if flag.override:
                    values = values[-flag.max_args:]
                else:
                    raise TooManyFlags(flag, values)

            # Special case:
            if flag.max_args == 1:
                value = await convert_flag(ctx, values[0], flag)
                setattr(self, flag.attribute, value)
                continue

            # Another special case, tuple parsing.
            # Tuple parsing is basically converting arguments within the flag
            # So, given flag: hello 20 as the input and Tuple[str, int] as the type hint
            # We would receive ('hello', 20) as the resulting value
            # This uses the same whitespace and quoting rules as regular parameters.
            values = [await convert_flag(ctx, value, flag) for value in values]

            if flag.cast_to_dict:
                values = dict(values)  # type: ignore

            setattr(self, flag.attribute, values)

        return self



from __future__ import annotations

import asyncio
import logging

import aiohttp

from .state import AutoShardedConnectionState
from .client import Client
from .backoff import ExponentialBackoff
from .gateway import *
from .error import (
    ClientException,
    HTTPException,
    GatewayNotFound,
    ConnectionClosed,
)

from typing import TYPE_CHECKING, Any, Callable, Tuple, Type, Optional, List, Dict, TypeVar

if TYPE_CHECKING:
    from .gateway import QQWebSocket

    EI = TypeVar('EI', bound='EventItem')

__all__ = (
    'AutoShardedClient',
    'ShardInfo',
)

_log = logging.getLogger(__name__)


class EventType:
    close = 0
    reconnect = 1
    resume = 2
    identify = 3
    terminate = 4
    clean_close = 5


class EventItem:
    __slots__ = ('type', 'shard', 'error')

    def __init__(self, etype: int, shard: Optional['Shard'], error: Optional[Exception]) -> None:
        self.type: int = etype
        self.shard: Optional['Shard'] = shard
        self.error: Optional[Exception] = error

    def __lt__(self: EI, other: EI) -> bool:
        if not isinstance(other, EventItem):
            return NotImplemented
        return self.type < other.type

    def __eq__(self: EI, other: EI) -> bool:
        if not isinstance(other, EventItem):
            return NotImplemented
        return self.type == other.type

    def __hash__(self) -> int:
        return hash(self.type)


class Shard:
    def __init__(self, ws: QQWebSocket, client: AutoShardedClient, queue_put: Callable[[EventItem], None]) -> None:
        self.ws: QQWebSocket = ws
        self._client: Client = client
        self._dispatch: Callable[..., None] = client.dispatch
        self._queue_put: Callable[[EventItem], None] = queue_put
        self.loop: asyncio.AbstractEventLoop = self._client.loop
        self._disconnect: bool = False
        self._reconnect = client._reconnect
        self._backoff: ExponentialBackoff = ExponentialBackoff()
        self._task: Optional[asyncio.Task] = None
        self._handled_exceptions: Tuple[Type[Exception], ...] = (
            OSError,
            HTTPException,
            GatewayNotFound,
            ConnectionClosed,
            aiohttp.ClientError,
            asyncio.TimeoutError,
        )

    @property
    def id(self) -> int:
        # QQWebSocket.shard_id is set in the from_client classmethod
        return self.ws.shard_id  # type: ignore

    def launch(self) -> None:
        self._task = self.loop.create_task(self.worker())

    def _cancel_task(self) -> None:
        if self._task is not None and not self._task.done():
            self._task.cancel()

    async def close(self) -> None:
        self._cancel_task()
        await self.ws.close(code=1000)

    async def disconnect(self) -> None:
        await self.close()
        self._dispatch('shard_disconnect', self.id)

    async def _handle_disconnect(self, e: Exception) -> None:
        self._dispatch('disconnect')
        self._dispatch('shard_disconnect', self.id)
        if not self._reconnect:
            self._queue_put(EventItem(EventType.close, self, e))
            return

        if self._client.is_closed():
            return

        if isinstance(e, OSError) and e.errno in (54, 10054):
            # If we get Connection reset by peer then always try to RESUME the connection.
            exc = ReconnectWebSocket(self.id, resume=True)
            self._queue_put(EventItem(EventType.resume, self, exc))
            return

        if isinstance(e, ConnectionClosed):
            if e.code != 1000:
                self._queue_put(EventItem(EventType.close, self, e))
                return

        retry = self._backoff.delay()
        _log.error('Attempting a reconnect for shard ID %s in %.2fs', self.id, retry, exc_info=e)
        await asyncio.sleep(retry)
        self._queue_put(EventItem(EventType.reconnect, self, e))

    async def worker(self) -> None:
        while not self._client.is_closed():
            try:
                await self.ws.poll_event()
            except ReconnectWebSocket as e:
                etype = EventType.resume if e.resume else EventType.identify
                self._queue_put(EventItem(etype, self, e))
                break
            except self._handled_exceptions as e:
                await self._handle_disconnect(e)
                break
            except asyncio.CancelledError:
                break
            except Exception as e:
                self._queue_put(EventItem(EventType.terminate, self, e))
                break

    async def reidentify(self, exc: ReconnectWebSocket) -> None:
        self._cancel_task()
        self._dispatch('disconnect')
        self._dispatch('shard_disconnect', self.id)
        _log.info('Got a request to %s the websocket at Shard ID %s.', exc.op, self.id)
        try:
            coro = QQWebSocket.from_client(
                self._client,
                resume=exc.resume,
                shard_id=self.id,
                session=self.ws.session_id,
                sequence=self.ws.sequence,
            )
            self.ws = await asyncio.wait_for(coro, timeout=60.0)
        except self._handled_exceptions as e:
            await self._handle_disconnect(e)
        except asyncio.CancelledError:
            return
        except Exception as e:
            self._queue_put(EventItem(EventType.terminate, self, e))
        else:
            self.launch()

    async def reconnect(self) -> None:
        self._cancel_task()
        try:
            coro = QQWebSocket.from_client(self._client, shard_id=self.id)
            self.ws = await asyncio.wait_for(coro, timeout=60.0)
        except self._handled_exceptions as e:
            await self._handle_disconnect(e)
        except asyncio.CancelledError:
            return
        except Exception as e:
            self._queue_put(EventItem(EventType.terminate, self, e))
        else:
            self.launch()


class ShardInfo:
    """提供信息和控制特定分片的类。
    你可以通过 :meth:`AutoShardedClient.get_shard` 或 :attr:`AutoShardedClient.shards` 检索这个对象。

    Attributes
    ------------
    id: :class:`int`
        此分片的分片 ID。
    shard_count: Optional[:class:`int`]
        此集群的分片计数。如果这是 ``None`` ，则机器人尚未启动。
    """

    __slots__ = ('_parent', 'id', 'shard_count')

    def __init__(self, parent: Shard, shard_count: Optional[int]) -> None:
        self._parent: Shard = parent
        self.id: int = parent.id
        self.shard_count: Optional[int] = shard_count

    def is_closed(self) -> bool:
        """:class:`bool`: 分片连接当前是否关闭。"""
        return not self._parent.ws.open

    async def disconnect(self) -> None:
        """|coro|
        断开分片。当这个被调用时，分片连接将关闭。
        如果分片已经断开连接，则不会执行任何操作。
        """
        if self.is_closed():
            return

        await self._parent.disconnect()

    async def reconnect(self) -> None:
        """|coro|
        断开连接，然后再次连接分片。
        """
        if not self.is_closed():
            await self._parent.disconnect()
        await self._parent.reconnect()

    async def connect(self) -> None:
        """|coro|
        连接一个分片。如果分片已连接，则不会执行任何操作。
        """
        if not self.is_closed():
            return

        await self._parent.reconnect()

    @property
    def latency(self) -> float:
        """:class:`float`: 测量此分片的 HEARTBEAT 和 HEARTBEAT_ACK 之间的延迟（以秒为单位）。"""
        return self._parent.ws.latency

    def is_ws_ratelimited(self) -> bool:
        """:class:`bool`: websocket 当前是否有速率限制。
        """
        return self._parent.ws.is_ratelimited()


class AutoShardedClient(Client):
    """一个类似于 :class:`Client` 的客户端，
    只是它将用户分片的复杂性处理成一个更易于管理和透明的单进程机器人。
    当使用这个客户端时，你将能够像 :class:`Client` 一样使用它，就像它是一个带有单个分片的常规客户端，当在内部实现时，它被分成多个分片。
    这使你不必处理 IPC 或其他复杂的基础。
    建议人数超过1000人以上才使用此客户端。
    如果没有提供 :attr:`.shard_count`，则库将使用 Bot Gateway 端点调用来确定要使用多少个分片。
    如果给出了 ``shard_ids`` 参数，则这些分片 ID 将用于启动内部分片。
    请注意，如果使用 :attr:`.shard_count` ，则必须提供。默认情况下，当省略时，客户端将启动从 0 到 ``shard_count - 1`` 的分片。

    Attributes
    ------------
    shard_ids: Optional[List[:class:`int`]]
        用于启动分片的可选 shard_id 列表。
    """

    if TYPE_CHECKING:
        _connection: AutoShardedConnectionState

    def __init__(self, *args: Any, loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs: Any) -> None:
        kwargs.pop('shard_id', None)
        self.shard_ids: Optional[List[int]] = kwargs.pop('shard_ids', None)
        super().__init__(*args, loop=loop, **kwargs)

        if self.shard_ids is not None:
            if self.shard_count is None:
                raise ClientException('在传递手动 shard_ids 时，你必须提供一个 shard_count。')
            elif not isinstance(self.shard_ids, (list, tuple)):
                raise ClientException('shard_ids 参数必须是列表或元组。')

        # instead of a single websocket, we have multiple
        # the key is the shard_id
        self.__shards = {}
        self._connection._get_websocket = self._get_websocket
        self._connection._get_client = lambda: self
        self.__queue = asyncio.PriorityQueue()

    def _get_websocket(self, guild_id: Optional[int] = None, *, shard_id: Optional[int] = None) -> QQWebSocket:
        if shard_id is None:
            # guild_id won't be None if shard_id is None and shard_count won't be None here
            shard_id = (guild_id >> 22) % self.shard_count  # type: ignore
        return self.__shards[shard_id].ws

    def _get_state(self, **options: Any) -> AutoShardedConnectionState:
        return AutoShardedConnectionState(
            dispatch=self.dispatch,
            handlers=self._handlers,
            hooks=self._hooks,
            http=self.http,
            loop=self.loop,
            **options,
        )

    @property
    def latency(self) -> float:
        """:class:`float`: 以秒为单位测量 HEARTBEAT 和 HEARTBEAT_ACK 之间的延迟。
        这与 :meth:`Client.latency` 的操作类似，但是它使用每个分片延迟的平均延迟。
        要获取分片延迟列表，请检查 :attr:`latencies` 属性。如果没有分片，则返回 ``nan`` 。
        """
        if not self.__shards:
            return float('nan')
        return sum(latency for _, latency in self.latencies) / len(self.__shards)

    @property
    def latencies(self) -> List[Tuple[int, float]]:
        """List[Tuple[:class:`int`, :class:`float`]]: HEARTBEAT 和 HEARTBEAT_ACK 之间的延迟列表（以秒为单位）。
        这将返回一个包含元素 ``(shard_id, latency)`` 的元组列表。
        """
        return [(shard_id, shard.ws.latency) for shard_id, shard in self.__shards.items()]

    def get_shard(self, shard_id: int) -> Optional[ShardInfo]:
        """Optional[:class:`ShardInfo`]: 获取给定分片 ID 的分片信息，如果未找到则为 ``None`` 。"""
        try:
            parent = self.__shards[shard_id]
        except KeyError:
            return None
        else:
            return ShardInfo(parent, self.shard_count)

    @property
    def shards(self) -> Dict[int, ShardInfo]:
        """Mapping[int, :class:`ShardInfo`]: 返回分片 ID 到其各自信息对象的映射。"""
        return {shard_id: ShardInfo(parent, self.shard_count) for shard_id, parent in self.__shards.items()}

    async def launch_shard(self, gateway: str, shard_id: int, *, initial: bool = False) -> None:
        try:
            coro = QQWebSocket.from_client(self, initial=initial, gateway=gateway, shard_id=shard_id)
            ws = await asyncio.wait_for(coro, timeout=180.0)
        except Exception:
            _log.exception('无法连接 shard_id: %s。正在重试...', shard_id)
            await asyncio.sleep(5.0)
            return await self.launch_shard(gateway, shard_id)

        # keep reading the shard while others connect
        self.__shards[shard_id] = ret = Shard(ws, self, self.__queue.put_nowait)
        ret.launch()

    async def launch_shards(self) -> None:
        if self.shard_count is None:
            self.shard_count, gateway = await self.http.get_bot_gateway()
        else:
            gateway = await self.http.get_gateway()

        self._connection.shard_count = self.shard_count

        shard_ids = self.shard_ids or range(self.shard_count)
        self._connection.shard_ids = shard_ids

        for shard_id in shard_ids:
            initial = shard_id == shard_ids[0]
            await self.launch_shard(gateway, shard_id, initial=initial)

        self._connection.shards_launched.set()

    async def connect(self, *, reconnect: bool = True) -> None:
        self._reconnect = reconnect
        await self.launch_shards()

        while not self.is_closed():
            item = await self.__queue.get()
            if item.type == EventType.close:
                await self.close()
                if isinstance(item.error, ConnectionClosed):
                    if item.error.code != 1000:
                        raise item.error
                return
            elif item.type in (EventType.identify, EventType.resume):
                await item.shard.reidentify(item.error)
            elif item.type == EventType.reconnect:
                await item.shard.reconnect()
            elif item.type == EventType.terminate:
                await self.close()
                raise item.error
            elif item.type == EventType.clean_close:
                return

    async def close(self) -> None:
        """|coro|
        关闭与 QQ 的连接。
        """
        if self.is_closed():
            return

        self._closed = True
        to_close = [asyncio.ensure_future(shard.close(), loop=self.loop) for shard in self.__shards.values()]
        if to_close:
            await asyncio.wait(to_close)

        await self.http.close()
        self.__queue.put_nowait(EventItem(EventType.clean_close, None, None))

    def is_ws_ratelimited(self) -> bool:
        """:class:`bool`: websocket 当前是否有速率限制。
        此实现检查是否有任何分片受速率限制。如需更精细的控制，请考虑 :meth:`ShardInfo.is_ws_ratelimited` 。
        """
        return any(shard.ws.is_ratelimited() for shard in self.__shards.values())



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Literal,
    List,
    Optional,
    Union,
    Set,
    Tuple,
    TypeVar,
    Type,
    TYPE_CHECKING,
    overload,
)
import asyncio
import functools
import inspect
import datetime

import qq

from .errors import *
from .cooldowns import Cooldown, BucketType, CooldownMapping, MaxConcurrency, DynamicCooldownMapping
from .converter import run_converters, get_converter, Greedy
from ._types import _BaseCommand
from .cog import Cog
from .context import Context

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec, TypeGuard

    from qq.message import Message

    from ._types import (
        Coro,
        CoroFunc,
        Check,
        Hook,
        Error,
    )

__all__ = (
    'Command',
    'Group',
    'GroupMixin',
    'command',
    'group',
    'has_role',
    'has_any_role',
    'check',
    'check_any',
    'before_invoke',
    'after_invoke',
    'bot_has_role',
    'bot_has_any_role',
    'cooldown',
    'dynamic_cooldown',
    'max_concurrency',
    'dm_only',
    'guild_only',
    'is_owner',
)

MISSING: Any = qq.utils.MISSING

T = TypeVar('T')
CogT = TypeVar('CogT', bound='Cog')
CommandT = TypeVar('CommandT', bound='Command')
ContextT = TypeVar('ContextT', bound='Context')
# CHT = TypeVar('CHT', bound='Check')
GroupT = TypeVar('GroupT', bound='Group')
HookT = TypeVar('HookT', bound='Hook')
ErrorT = TypeVar('ErrorT', bound='Error')

if TYPE_CHECKING:
    P = ParamSpec('P')
else:
    P = TypeVar('P')


def unwrap_function(function: Callable[..., Any]) -> Callable[..., Any]:
    partial = functools.partial
    while True:
        if hasattr(function, '__wrapped__'):
            function = function.__wrapped__
        elif isinstance(function, partial):
            function = function.func
        else:
            return function


def get_signature_parameters(function: Callable[..., Any], globalns: Dict[str, Any]) -> Dict[str, inspect.Parameter]:
    signature = inspect.signature(function)
    params = {}
    cache: Dict[str, Any] = {}
    eval_annotation = qq.utils.evaluate_annotation
    for name, parameter in signature.parameters.items():
        annotation = parameter.annotation
        if annotation is parameter.empty:
            params[name] = parameter
            continue
        if annotation is None:
            params[name] = parameter.replace(annotation=type(None))
            continue

        annotation = eval_annotation(annotation, globalns, globalns, cache)
        if annotation is Greedy:
            raise TypeError('Unparameterized Greedy[...] is disallowed in signature.')

        params[name] = parameter.replace(annotation=annotation)

    return params


def wrap_callback(coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        try:
            ret = await coro(*args, **kwargs)
        except CommandError:
            raise
        except asyncio.CancelledError:
            return
        except Exception as exc:
            raise CommandInvokeError(exc) from exc
        return ret

    return wrapped


def hooked_wrapped_callback(command, ctx, coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        try:
            ret = await coro(*args, **kwargs)
        except CommandError:
            ctx.command_failed = True
            raise
        except asyncio.CancelledError:
            ctx.command_failed = True
            return
        except Exception as exc:
            ctx.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if command._max_concurrency is not None:
                await command._max_concurrency.release(ctx)

            await command.call_after_hooks(ctx)
        return ret

    return wrapped


class _CaseInsensitiveDict(dict):
    def __contains__(self, k):
        return super().__contains__(k.casefold())

    def __delitem__(self, k):
        return super().__delitem__(k.casefold())

    def __getitem__(self, k):
        return super().__getitem__(k.casefold())

    def get(self, k, default=None):
        return super().get(k.casefold(), default)

    def pop(self, k, default=None):
        return super().pop(k.casefold(), default)

    def __setitem__(self, k, v):
        super().__setitem__(k.casefold(), v)


class Command(_BaseCommand, Generic[CogT, P, T]):
    r"""一个实现机器人文本命令协议的类。

    这些不是手动创建的，而是通过装饰器或功能接口创建的。

    Attributes
    -----------
    name: :class:`str`
        命令的名称。
    callback: :ref:`coroutine <coroutine>`
        调用命令时执行的协程。
    help: Optional[:class:`str`]
        命令的长帮助文本。
    brief: Optional[:class:`str`]
        命令的简短帮助文本。
    usage: Optional[:class:`str`]
        替换默认帮助文本中的参数。
    aliases: Union[List[:class:`str`], Tuple[:class:`str`]]
        可以在其下调用命令的别名列表。
    enabled: :class:`bool`
        指示当前是否启用命令的布尔值。
        如果命令在禁用时被调用，则 :exc:`.DisabledCommand` 将引发 :func:`.on_command_error` 事件。默认为 ``True`` 。
    parent: Optional[:class:`Group`]
        此命令所属的父组。 ``None`` 如果没有的话。
    cog: Optional[:class:`Cog`]
        该命令所属的齿轮。 ``None`` 如果没有的话。
    checks: List[Callable[[:class:`.Context`], :class:`bool`]]
        一个谓词列表，用于验证是否可以使用给定的 :class:`.Context` 作为唯一参数来执行命令。
         如果必须抛出异常以表示失败，则应使用继承自 :exc:`.CommandError` 的异常。
         请注意，如果检查失败，则 :exc:`.CheckFailure` 异常将引发到 :func:`.on_command_error` 事件。
    description: :class:`str`
        默认帮助命令中带有前缀的消息。
    hidden: :class:`bool`
        如果为  ``True``\ ，则默认帮助命令不会在帮助输出中显示此内容。
    rest_is_raw: :class:`bool`
        如果 ``False`` 则仅关键字参数，则仅关键字参数将被剥离并处理，
        就好像它是处理 :exc:`.MissingRequiredArgument` 和默认值的常规参数一样，
        而不是传递原始数据。 如果 ``True`` 则仅关键字参数将以完全原始的方式传递其余参数。 默认为 ``False`` 。
    invoked_subcommand: Optional[:class:`Command`]
        调用的子命令（如果有）。
    require_var_positional: :class:`bool`
        如果 ``True`` 并且指定了可变参数位置参数，则要求用户至少指定一个参数。 默认为 ``False`` 。

    ignore_extra: :class:`bool`
        如果 ``True``\，如果它的所有要求都得到满足，则忽略传递给命令的无关字符串
        （例如 ``?foo a b c`` 当只需要 ``a`` 和 ``b`` 时）。
        否则 :func:`.on_command_error` 和本地错误处理程序使用 :exc:`.TooManyArguments` 调用。 默认为 ``True`` 。
    cooldown_after_parsing: :class:`bool`
        如果为 ``True``\，则在参数解析后完成冷却处理，这会调用转换器。 如果 ``False`` 则首先完成冷却处理，然后再调用转换器。 默认为 ``False``  。
    extras: :class:`dict`
        用户的字典提供了附加到命令的附加内容。
        
        .. note::
            该对象可由库复制。

    """
    __original_kwargs__: Dict[str, Any]

    def __new__(cls: Type[CommandT], *args: Any, **kwargs: Any) -> CommandT:
        # if you're wondering why this is done, it's because we need to ensure
        # we have a complete original copy of **kwargs even for classes that
        # mess with it by popping before delegating to the subclass __init__.
        # In order to do this, we need to control the instance creation and
        # inject the original kwargs through __new__ rather than doing it
        # inside __init__.
        self = super().__new__(cls)

        # we do a shallow copy because it's probably the most common use case.
        # this could potentially break if someone modifies a list or something
        # while it's in movement, but for now this is the cheapest and
        # fastest way to do what we want.
        self.__original_kwargs__ = kwargs.copy()
        return self

    def __init__(self, func: Union[
        Callable[Concatenate[CogT, ContextT, P], Coro[T]],
        Callable[Concatenate[ContextT, P], Coro[T]],
    ], **kwargs: Any):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('回调必须是协程。')

        name = kwargs.get('name') or func.__name__
        if not isinstance(name, str):
            raise TypeError('命令的名称必须是字符串。')
        self.name: str = name

        self.callback = func
        self.enabled: bool = kwargs.get('enabled', True)

        help_doc = kwargs.get('help')
        if help_doc is not None:
            help_doc = inspect.cleandoc(help_doc)
        else:
            help_doc = inspect.getdoc(func)
            if isinstance(help_doc, bytes):
                help_doc = help_doc.decode('utf-8')

        self.help: Optional[str] = help_doc

        self.brief: Optional[str] = kwargs.get('brief')
        self.usage: Optional[str] = kwargs.get('usage')
        self.rest_is_raw: bool = kwargs.get('rest_is_raw', False)
        self.aliases: Union[List[str], Tuple[str]] = kwargs.get('aliases', [])
        self.extras: Dict[str, Any] = kwargs.get('extras', {})

        if not isinstance(self.aliases, (list, tuple)):
            raise TypeError("命令的别名必须是一个列表或一个字符串元组。")

        self.description: str = inspect.cleandoc(kwargs.get('description', ''))
        self.hidden: bool = kwargs.get('hidden', False)

        try:
            checks = func.__commands_checks__
            checks.reverse()
        except AttributeError:
            checks = kwargs.get('checks', [])

        self.checks: List[Check] = checks

        try:
            cooldown = func.__commands_cooldown__
        except AttributeError:
            cooldown = kwargs.get('cooldown')

        if cooldown is None:
            buckets = CooldownMapping(cooldown, BucketType.default)
        elif isinstance(cooldown, CooldownMapping):
            buckets = cooldown
        else:
            raise TypeError("Cooldown 必须是 CooldownMapping 或 None 的一个实例。")
        self._buckets: CooldownMapping = buckets

        try:
            max_concurrency = func.__commands_max_concurrency__
        except AttributeError:
            max_concurrency = kwargs.get('max_concurrency')

        self._max_concurrency: Optional[MaxConcurrency] = max_concurrency

        self.require_var_positional: bool = kwargs.get('require_var_positional', False)
        self.ignore_extra: bool = kwargs.get('ignore_extra', True)
        self.cooldown_after_parsing: bool = kwargs.get('cooldown_after_parsing', False)
        self.cog: Optional[CogT] = None

        # bandaid for the fact that sometimes parent can be the bot instance
        parent = kwargs.get('parent')
        self.parent: Optional[GroupMixin] = parent if isinstance(parent, _BaseCommand) else None  # type: ignore

        self._before_invoke: Optional[Hook] = None
        try:
            before_invoke = func.__before_invoke__
        except AttributeError:
            pass
        else:
            self.before_invoke(before_invoke)

        self._after_invoke: Optional[Hook] = None
        try:
            after_invoke = func.__after_invoke__
        except AttributeError:
            pass
        else:
            self.after_invoke(after_invoke)

    @property
    def callback(self) -> Union[
        Callable[Concatenate[CogT, Context, P], Coro[T]],
        Callable[Concatenate[Context, P], Coro[T]],
    ]:
        return self._callback

    @callback.setter
    def callback(self, function: Union[
        Callable[Concatenate[CogT, Context, P], Coro[T]],
        Callable[Concatenate[Context, P], Coro[T]],
    ]) -> None:
        self._callback = function
        unwrap = unwrap_function(function)
        self.module = unwrap.__module__

        try:
            globalns = unwrap.__globals__
        except AttributeError:
            globalns = {}

        self.params = get_signature_parameters(function, globalns)

    def add_check(self, func: Check) -> None:
        """向命令添加检查。

        这是 :func:`.check` 的非装饰器接口。


        Parameters
        -----------
        func
            将用作检查的函数。
        """

        self.checks.append(func)

    def remove_check(self, func: Check) -> None:
        """从命令中删除检查。

        此函数是幂等的，如果该函数不在命令的检查中，则不会引发异常。

        Parameters
        -----------
        func
            要从检查中删除的函数。
        """

        try:
            self.checks.remove(func)
        except ValueError:
            pass

    def update(self, **kwargs: Any) -> None:
        """使用更新的属性更新 :class:`Command` 实例。

        这在参数方面与 :func:`.command` 装饰器类似，因为它们被传递给 :class:`Command` 或子类构造函数，没有名称和回调。
        """
        self.__init__(self.callback, **dict(self.__original_kwargs__, **kwargs))

    async def __call__(self, context: Context, *args: P.args, **kwargs: P.kwargs) -> T:
        """|coro|

        调用命令持有的内部回调。

        .. note::

            这绕过了所有机制——包括检查、转换器、调用钩子、冷却等。你必须小心地将正确的参数和类型传递给这个函数。
        """
        if self.cog is not None:
            return await self.callback(self.cog, context, *args, **kwargs)  # type: ignore
        else:
            return await self.callback(context, *args, **kwargs)  # type: ignore

    def _ensure_assignment_on_copy(self, other: CommandT) -> CommandT:
        other._before_invoke = self._before_invoke
        other._after_invoke = self._after_invoke
        if self.checks != other.checks:
            other.checks = self.checks.copy()
        if self._buckets.valid and not other._buckets.valid:
            other._buckets = self._buckets.copy()
        if self._max_concurrency != other._max_concurrency:
            # _max_concurrency won't be None at this point
            other._max_concurrency = self._max_concurrency.copy()  # type: ignore

        try:
            other.on_error = self.on_error
        except AttributeError:
            pass
        return other

    def copy(self: CommandT) -> CommandT:
        """创建此命令的副本。

        Returns
        --------
        :class:`Command`
            此命令的新实例。
        """
        ret = self.__class__(self.callback, **self.__original_kwargs__)
        return self._ensure_assignment_on_copy(ret)

    def _update_copy(self: CommandT, kwargs: Dict[str, Any]) -> CommandT:
        if kwargs:
            kw = kwargs.copy()
            kw.update(self.__original_kwargs__)
            copy = self.__class__(self.callback, **kw)
            return self._ensure_assignment_on_copy(copy)
        else:
            return self.copy()

    async def dispatch_error(self, ctx: Context, error: Exception) -> None:
        ctx.command_failed = True
        cog = self.cog
        try:
            coro = self.on_error
        except AttributeError:
            pass
        else:
            injected = wrap_callback(coro)
            if cog is not None:
                await injected(cog, ctx, error)
            else:
                await injected(ctx, error)

        try:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_command_error)
                if local is not None:
                    wrapped = wrap_callback(local)
                    await wrapped(ctx, error)
        finally:
            ctx.bot.dispatch('command_error', ctx, error)

    async def transform(self, ctx: Context, param: inspect.Parameter) -> Any:
        required = param.default is param.empty
        converter = get_converter(param)
        consume_rest_is_special = param.kind == param.KEYWORD_ONLY and not self.rest_is_raw
        view = ctx.view
        view.skip_ws()

        # The greedy converter is simple -- it keeps going until it fails in which case,
        # it undos the view ready for the next parameter to use instead
        if isinstance(converter, Greedy):
            if param.kind in (param.POSITIONAL_OR_KEYWORD, param.POSITIONAL_ONLY):
                return await self._transform_greedy_pos(ctx, param, required, converter.converter)
            elif param.kind == param.VAR_POSITIONAL:
                return await self._transform_greedy_var_pos(ctx, param, converter.converter)
            else:
                # if we're here, then it's a KEYWORD_ONLY param type
                # since this is mostly useless, we'll helpfully transform Greedy[X]
                # into just X and do the parsing that way.
                converter = converter.converter

        if view.eof:
            if param.kind == param.VAR_POSITIONAL:
                raise RuntimeError()  # break the loop
            if required:
                if self._is_typing_optional(param.annotation):
                    return None
                if hasattr(converter, '__commands_is_flag__') and converter._can_be_constructible():
                    return await converter._construct_default(ctx)
                raise MissingRequiredArgument(param)
            return param.default

        previous = view.index
        if consume_rest_is_special:
            argument = view.read_rest().strip()
        else:
            try:
                argument = view.get_quoted_word()
            except ArgumentParsingError as exc:
                if self._is_typing_optional(param.annotation):
                    view.index = previous
                    return None
                else:
                    raise exc
        view.previous = previous

        # type-checker fails to narrow argument
        return await run_converters(ctx, converter, argument, param)  # type: ignore

    async def _transform_greedy_pos(self, ctx: Context, param: inspect.Parameter, required: bool,
                                    converter: Any) -> Any:
        view = ctx.view
        result = []
        while not view.eof:
            # for use with a manual undo
            previous = view.index

            view.skip_ws()
            try:
                argument = view.get_quoted_word()
                value = await run_converters(ctx, converter, argument, param)  # type: ignore
            except (CommandError, ArgumentParsingError):
                view.index = previous
                break
            else:
                result.append(value)

        if not result and not required:
            return param.default
        return result

    async def _transform_greedy_var_pos(self, ctx: Context, param: inspect.Parameter, converter: Any) -> Any:
        view = ctx.view
        previous = view.index
        try:
            argument = view.get_quoted_word()
            value = await run_converters(ctx, converter, argument, param)  # type: ignore
        except (CommandError, ArgumentParsingError):
            view.index = previous
            raise RuntimeError() from None  # break loop
        else:
            return value

    @property
    def clean_params(self) -> Dict[str, inspect.Parameter]:
        """Dict[:class:`str`, :class:`inspect.Parameter`]:
        检索没有上下文或自身参数的参数字典。

        用于检查签名。
        """
        result = self.params.copy()
        if self.cog is not None:
            # first parameter is self
            try:
                del result[next(iter(result))]
            except StopIteration:
                raise ValueError("缺少 'self' 参数") from None

        try:
            # first/second parameter is context
            del result[next(iter(result))]
        except StopIteration:
            raise ValueError("缺少 'context' 参数") from None

        return result

    @property
    def full_parent_name(self) -> str:
        """:class:`str`: 检索完全限定的父命令名称。

        这是执行它所需的基本命令名称。 例如，在 ``?one two three``  中，父名称将是 ``one two``  。
        """
        entries = []
        command = self
        # command.parent is type-hinted as GroupMixin some attributes are resolved via MRO
        while command.parent is not None:  # type: ignore
            command = command.parent  # type: ignore
            entries.append(command.name)  # type: ignore

        return ' '.join(reversed(entries))

    @property
    def parents(self) -> List[Group]:
        """List[:class:`Group`]: 检索此命令的父级。

        如果该命令没有父级，则它返回一个空的 :class:`list`。

        例如在命令 ``?a b c test`` 中，父级是 ``[c, b, a]``。
        """
        entries = []
        command = self
        while command.parent is not None:  # type: ignore
            command = command.parent  # type: ignore
            entries.append(command)

        return entries

    @property
    def root_parent(self) -> Optional[Group]:
        """Optional[:class:`Group`]: 检索此命令的根父级。

        如果该命令没有父级，则返回 ``None`` 。

        例如在命令 ``?a b c test`` 中，根父节点是 ``a`` 。
        """
        if not self.parent:
            return None
        return self.parents[-1]

    @property
    def qualified_name(self) -> str:
        """:class:`str`: 检索完全限定的命令名称。

        这也是带有命令名称的完整父名称。 例如，在 ``?one two three`` 中，限定名称将是 ``one two three`` 。
        """

        parent = self.full_parent_name
        if parent:
            return parent + ' ' + self.name
        else:
            return self.name

    def __str__(self) -> str:
        return self.qualified_name

    async def _parse_arguments(self, ctx: Context) -> None:
        ctx.args = [ctx] if self.cog is None else [self.cog, ctx]
        ctx.kwargs = {}
        args = ctx.args
        kwargs = ctx.kwargs

        view = ctx.view
        iterator = iter(self.params.items())

        if self.cog is not None:
            # we have 'self' as the first parameter so just advance
            # the iterator and resume parsing
            try:
                next(iterator)
            except StopIteration:
                raise qq.ClientException(f'{self.name} 命令的回调缺少 "self" 参数。')

        # next we have the 'ctx' as the next parameter
        try:
            next(iterator)
        except StopIteration:
            raise qq.ClientException(f'{self.name} 命令的回调缺少 "ctx" 参数。')

        for name, param in iterator:
            ctx.current_parameter = param
            if param.kind in (param.POSITIONAL_OR_KEYWORD, param.POSITIONAL_ONLY):
                transformed = await self.transform(ctx, param)
                args.append(transformed)
            elif param.kind == param.KEYWORD_ONLY:
                # kwarg only param denotes "consume rest" semantics
                if self.rest_is_raw:
                    converter = get_converter(param)
                    argument = view.read_rest()
                    kwargs[name] = await run_converters(ctx, converter, argument, param)
                else:
                    kwargs[name] = await self.transform(ctx, param)
                break
            elif param.kind == param.VAR_POSITIONAL:
                if view.eof and self.require_var_positional:
                    raise MissingRequiredArgument(param)
                while not view.eof:
                    try:
                        transformed = await self.transform(ctx, param)
                        args.append(transformed)
                    except RuntimeError:
                        break

        if not self.ignore_extra and not view.eof:
            raise TooManyArguments(f'传递给{self.qualified_name}的参数太多')

    async def call_before_hooks(self, ctx: Context) -> None:
        # now that we're done preparing we can call the pre-command hooks
        # first, call the command local hook:
        cog = self.cog
        if self._before_invoke is not None:
            # should be cog if @commands.before_invoke is used
            instance = getattr(self._before_invoke, '__self__', cog)
            # __self__ only exists for methods, not functions
            # however, if @command.before_invoke is used, it will be a function
            if instance:
                await self._before_invoke(instance, ctx)  # type: ignore
            else:
                await self._before_invoke(ctx)  # type: ignore

        # call the cog local hook if applicable:
        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_before_invoke)
            if hook is not None:
                await hook(ctx)

        # call the bot global hook if necessary
        hook = ctx.bot._before_invoke
        if hook is not None:
            await hook(ctx)

    async def call_after_hooks(self, ctx: Context) -> None:
        cog = self.cog
        if self._after_invoke is not None:
            instance = getattr(self._after_invoke, '__self__', cog)
            if instance:
                await self._after_invoke(instance, ctx)  # type: ignore
            else:
                await self._after_invoke(ctx)  # type: ignore

        # call the cog local hook if applicable:
        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_after_invoke)
            if hook is not None:
                await hook(ctx)

        hook = ctx.bot._after_invoke
        if hook is not None:
            await hook(ctx)

    def _prepare_cooldowns(self, ctx: Context) -> None:
        if self._buckets.valid:
            dt = ctx.message.edited_at or ctx.message.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(ctx.message, current)
            if bucket is not None:
                retry_after = bucket.update_rate_limit(current)
                if retry_after:
                    raise CommandOnCooldown(bucket, retry_after, self._buckets.type)  # type: ignore

    async def prepare(self, ctx: Context) -> None:
        ctx.command = self

        if not await self.can_run(ctx):
            raise CheckFailure(f'命令 {self.qualified_name} 的检查函数失败。')

        if self._max_concurrency is not None:
            # For this application, context can be duck-typed as a Message
            await self._max_concurrency.acquire(ctx)  # type: ignore

        try:
            if self.cooldown_after_parsing:
                await self._parse_arguments(ctx)
                self._prepare_cooldowns(ctx)
            else:
                self._prepare_cooldowns(ctx)
                await self._parse_arguments(ctx)

            await self.call_before_hooks(ctx)
        except:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(ctx)  # type: ignore
            raise

    def is_on_cooldown(self, ctx: Context) -> bool:
        """检查命令当前是否处于冷却状态。

        Parameters
        -----------
        ctx: :class:`.Context`
            检查命令冷却状态时使用的调用上下文。

        Returns
        --------
        :class:`bool`
            指示命令是否处于冷却状态的布尔值。
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(ctx.message)
        dt = ctx.message.edited_at or ctx.message.created_at
        current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        return bucket.get_tokens(current) == 0

    def reset_cooldown(self, ctx: Context) -> None:
        """重置此命令的冷却时间。

        Parameters
        -----------
        ctx: :class:`.Context`
            重置冷却时间的调用上下文。
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            bucket.reset()

    def get_cooldown_retry_after(self, ctx: Context) -> float:
        """检索可以再次尝试此命令之前的秒数。

        Parameters
        -----------
        ctx: :class:`.Context`
            从中检索冷却时间的调用上下文。

        Returns
        --------
        :class:`float`
            此命令的冷却剩余时间（以秒为单位）。
            如果这是 ``0.0`` ，则该命令不在冷却中。
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            dt = ctx.message.edited_at or ctx.message.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            return bucket.get_retry_after(current)

        return 0.0

    async def invoke(self, ctx: Context) -> None:
        await self.prepare(ctx)

        # terminate the invoked_subcommand chain.
        # since we're in a regular command (and not a group) then
        # the invoked subcommand is None.
        ctx.invoked_subcommand = None
        ctx.subcommand_passed = None
        injected = hooked_wrapped_callback(self, ctx, self.callback)
        await injected(*ctx.args, **ctx.kwargs)

    async def reinvoke(self, ctx: Context, *, call_hooks: bool = False) -> None:
        ctx.command = self
        await self._parse_arguments(ctx)

        if call_hooks:
            await self.call_before_hooks(ctx)

        ctx.invoked_subcommand = None
        try:
            await self.callback(*ctx.args, **ctx.kwargs)  # type: ignore
        except:
            ctx.command_failed = True
            raise
        finally:
            if call_hooks:
                await self.call_after_hooks(ctx)

    def error(self, coro: ErrorT) -> ErrorT:
        """将协程注册为本地错误处理程序的装饰器。

        本地错误处理程序是一个仅限于单个命令的 :func:`.on_command_error` 事件。
        然而， :func:`.on_command_error` 之后仍然会被调用。

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            要注册为本地错误处理程序的协程。

        Raises
        -------
        TypeError
            传递的协程实际上并不是协程。
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The error handler must be a coroutine.')

        self.on_error: Error = coro
        return coro

    def has_error_handler(self) -> bool:
        """:class:`bool`: 检查命令是否已注册错误处理程序。
        """
        return hasattr(self, 'on_error')

    def before_invoke(self, coro: HookT) -> HookT:
        """将协程注册为调用前钩的装饰器。

        在调用命令之前直接调用调用前钩。 这使得它成为设置数据库连接或所需的任何类型的设置的有用功能。

        这个调用前钩有一个唯一的参数，一个 :class:`.Context`。

        有关更多信息，请参阅 :meth:`.Bot.before_invoke`。

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            要注册为调用前钩的协程。

        Raises
        -------
        TypeError
            传递的协程实际上并不是协程。
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('调用前钩必须是一个协程。')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: HookT) -> HookT:
        """将协程注册为调用后钩的装饰器。

        在调用命令后直接调用后钩。 这使其成为清理数据库连接或任何类型的清理所需的有用功能。

        这个调用后钩有一个唯一的参数，一个 :class:`.Context`。

        有关更多信息，请参阅 :meth:`.Bot.after_invoke`。

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            要注册为调用后钩的协程。

        Raises
        -------
        TypeError
            传递的协程实际上并不是协程。
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('调用后钩必须是一个协程。')

        self._after_invoke = coro
        return coro

    @property
    def cog_name(self) -> Optional[str]:
        """Optional[:class:`str`]: 此命令所属的 cog 的名称（如果有）。"""
        return type(self.cog).__cog_name__ if self.cog is not None else None

    @property
    def short_doc(self) -> str:
        """:class:`str`: 获取命令的 ``简短`` 文档。

        默认情况下，这是 :attr:`.brief` 属性。
        如果该查找导致空字符串，则使用 :attr:`.help` 属性的第一行。
        """
        if self.brief is not None:
            return self.brief
        if self.help is not None:
            return self.help.split('\n', 1)[0]
        return ''

    def _is_typing_optional(self, annotation: Union[T, Optional[T]]) -> TypeGuard[Optional[T]]:
        return getattr(annotation, '__origin__', None) is Union and type(None) in annotation.__args__  # type: ignore

    @property
    def signature(self) -> str:
        """:class:`str`: 返回对帮助命令输出有用的类似 POSIX 的签名。"""
        if self.usage is not None:
            return self.usage

        params = self.clean_params
        if not params:
            return ''

        result = []
        for name, param in params.items():
            greedy = isinstance(param.annotation, Greedy)
            optional = False  # postpone evaluation of if it's an optional argument

            # for typing.Literal[...], typing.Optional[typing.Literal[...]], and Greedy[typing.Literal[...]], the
            # parameter signature is a literal list of it's values
            annotation = param.annotation.converter if greedy else param.annotation
            origin = getattr(annotation, '__origin__', None)
            if not greedy and origin is Union:
                none_cls = type(None)
                union_args = annotation.__args__
                optional = union_args[-1] is none_cls
                if len(union_args) == 2 and optional:
                    annotation = union_args[0]
                    origin = getattr(annotation, '__origin__', None)

            if origin is Literal:
                name = '|'.join(f'"{v}"' if isinstance(v, str) else str(v) for v in annotation.__args__)
            if param.default is not param.empty:
                # We don't want None or '' to trigger the [name=value] case and instead it should
                # do [name] since [name=None] or [name=] are not exactly useful for the user.
                should_print = param.default if isinstance(param.default, str) else param.default is not None
                if should_print:
                    result.append(f'[{name}={param.default}]' if not greedy else
                                  f'[{name}={param.default}]...')
                    continue
                else:
                    result.append(f'[{name}]')

            elif param.kind == param.VAR_POSITIONAL:
                if self.require_var_positional:
                    result.append(f'<{name}...>')
                else:
                    result.append(f'[{name}...]')
            elif greedy:
                result.append(f'[{name}]...')
            elif optional:
                result.append(f'[{name}]')
            else:
                result.append(f'<{name}>')

        return ' '.join(result)

    async def can_run(self, ctx: Context) -> bool:
        """|coro|

        通过检查 :attr:`~Command.checks` 属性中的所有谓词来检查命令是否可以执行。 这还会检查命令是否被禁用。

        Parameters
        -----------
        ctx: :class:`.Context`
            当前正在调用的命令的 ctx。

        Raises
        -------
        :class:`CommandError`
            在检查调用期间引发的任何命令错误都将由此函数传播。

        Returns
        --------
        :class:`bool`
            指示是否可以调用命令的布尔值。
        """

        if not self.enabled:
            raise DisabledCommand(f'{self.name} 命令被禁用')

        original = ctx.command
        ctx.command = self

        try:
            if not await ctx.bot.can_run(ctx):
                raise CheckFailure(f'命令 {self.qualified_name} 的全局检查函数失败。')

            cog = self.cog
            if cog is not None:
                local_check = Cog._get_overridden_method(cog.cog_check)
                if local_check is not None:
                    ret = await qq.utils.maybe_coroutine(local_check, ctx)
                    if not ret:
                        return False

            predicates = self.checks
            if not predicates:
                # since we have no checks, then we just return True.
                return True

            return await qq.utils.async_all(predicate(ctx) for predicate in predicates)  # type: ignore
        finally:
            ctx.command = original


class GroupMixin(Generic[CogT]):
    """一个 mixin，它为行为类似于 :class:`.Group` 的类实现通用功能，并允许注册命令。

    Attributes
    -----------
    all_commands: :class:`dict`
        命令名称到 :class:`.Command` 对象的映射。
    case_insensitive: :class:`bool`
        命令是否应该不区分大小写。 默认为 ``False`` 。
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        case_insensitive = kwargs.get('case_insensitive', False)
        self.all_commands: Dict[str, Command[CogT, Any, Any]] = _CaseInsensitiveDict() if case_insensitive else {}
        self.case_insensitive: bool = case_insensitive
        super().__init__(*args, **kwargs)

    @property
    def commands(self) -> Set[Command[CogT, Any, Any]]:
        """Set[:class:`.Command`]: 一组独特的没有注册别名的命令。"""
        return set(self.all_commands.values())

    def recursively_remove_all_commands(self) -> None:
        for command in self.all_commands.copy().values():
            if isinstance(command, GroupMixin):
                command.recursively_remove_all_commands()
            self.remove_command(command.name)

    def add_command(self, command: Command[CogT, Any, Any]) -> None:
        """将 :class:`.Command` 添加到内部命令列表中。

        这通常不被调用，而是使用 :meth:`~.GroupMixin.command` 或 :meth:`~.GroupMixin.group` 快捷装饰器。

        Parameters
        -----------
        command: :class:`Command`
            要添加的命令。

        Raises
        -------
        :exc:`.CommandRegistrationError`
            如果该命令或其别名已被不同的命令注册。
        TypeError
            如果传递的命令不是 :class:`.Command` 的子类。
        """

        if not isinstance(command, Command):
            raise TypeError('传递的命令必须是 Command 的子类')

        if isinstance(self, Command):
            command.parent = self

        if command.name in self.all_commands:
            raise CommandRegistrationError(command.name)

        self.all_commands[command.name] = command
        for alias in command.aliases:
            if alias in self.all_commands:
                self.remove_command(command.name)
                raise CommandRegistrationError(alias, alias_conflict=True)
            self.all_commands[alias] = command

    def remove_command(self, name: str) -> Optional[Command[CogT, Any, Any]]:
        """从内部命令列表中删除 :class:`.Command`。

        这也可以用作删除别名的方法。

        Parameters
        -----------
        name: :class:`str`
            要删除的命令的名称。

        Returns
        --------
        Optional[:class:`.Command`]
            删除的命令。 如果名称无效，则返回 ``None`` 。
        """
        command = self.all_commands.pop(name, None)

        # does not exist
        if command is None:
            return None

        if name in command.aliases:
            # we're removing an alias so we don't want to remove the rest
            return command

        # we're not removing the alias so let's delete the rest of them.
        for alias in command.aliases:
            cmd = self.all_commands.pop(alias, None)
            # in the case of a CommandRegistrationError, an alias might conflict
            # with an already existing command. If this is the case, we want to
            # make sure the pre-existing command is not removed.
            if cmd is not None and cmd != command:
                self.all_commands[alias] = cmd
        return command

    def walk_commands(self) -> Generator[Command[CogT, Any, Any], None, None]:
        """递归遍历所有命令和子命令的迭代器。

        Yields
        ------
        Union[:class:`.Command`, :class:`.Group`]
            来自内部命令列表的命令或组。
        """
        for command in self.commands:
            yield command
            if isinstance(command, GroupMixin):
                yield from command.walk_commands()

    def get_command(self, name: str) -> Optional[Command[CogT, Any, Any]]:
        """从内部命令列表中获取 :class:`.Command` 。

        这也可以用作获取别名的一种方式。

        该名称可以是完全限定的（例如 ``foo bar`` ）将获得组命令 ``foo`` 的子命令 ``bar`` 。
        如果未找到子命令，则像往常一样返回 ``None`` 。

        Parameters
        -----------
        name: :class:`str`
            要获取的命令的名称。

        Returns
        --------
        Optional[:class:`Command`]
            请求的命令。 如果未找到，则返回 ``None`` 。
        """

        # fast path, no space in name.
        if ' ' not in name:
            return self.all_commands.get(name)

        names = name.split()
        if not names:
            return None
        obj = self.all_commands.get(names[0])
        if not isinstance(obj, GroupMixin):
            return obj

        for name in names[1:]:
            try:
                obj = obj.all_commands[name]  # type: ignore
            except (AttributeError, KeyError):
                return None

        return obj

    @overload
    def command(
            self,
            name: str = ...,
            cls: Type[Command[CogT, P, T]] = ...,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[
        [
            Union[
                Callable[Concatenate[CogT, ContextT, P], Coro[T]],
                Callable[Concatenate[ContextT, P], Coro[T]],
            ]
        ], Command[CogT, P, T]]:
        ...

    @overload
    def command(
            self,
            name: str = ...,
            cls: Type[CommandT] = ...,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[[Callable[Concatenate[ContextT, P], Coro[Any]]], CommandT]:
        ...

    def command(
            self,
            name: str = MISSING,
            cls: Type[CommandT] = MISSING,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[[Callable[Concatenate[ContextT, P], Coro[Any]]], CommandT]:
        """调用 :func:`.command` 并通过 :meth:`~.GroupMixin.add_command` 将其添加到内部命令列表的快捷方式装饰器。

        Returns
        --------
        Callable[..., :class:`Command`]
            将提供的方法转换为命令的装饰器，将其添加到机器人，然后返回它。
        """

        def decorator(func: Callable[Concatenate[ContextT, P], Coro[Any]]) -> CommandT:
            kwargs.setdefault('parent', self)
            result = command(name=name, cls=cls, *args, **kwargs)(func)
            self.add_command(result)
            return result

        return decorator

    @overload
    def group(
            self,
            name: str = ...,
            cls: Type[Group[CogT, P, T]] = ...,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[[
                      Union[
                          Callable[Concatenate[CogT, ContextT, P], Coro[T]],
                          Callable[Concatenate[ContextT, P], Coro[T]]
                      ]
                  ], Group[CogT, P, T]]:
        ...

    @overload
    def group(
            self,
            name: str = ...,
            cls: Type[GroupT] = ...,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[[Callable[Concatenate[ContextT, P], Coro[Any]]], GroupT]:
        ...

    def group(
            self,
            name: str = MISSING,
            cls: Type[GroupT] = MISSING,
            *args: Any,
            **kwargs: Any,
    ) -> Callable[[Callable[Concatenate[ContextT, P], Coro[Any]]], GroupT]:
        """调用 :func:`.group` 并通过 :meth:`~.GroupMixin.add_command` 将其添加到内部命令列表的快捷方式装饰器。

        Returns
        --------
        Callable[..., :class:`Group`]
            将提供的方法转换为 Group 的装饰器，将其添加到机器人，然后返回它。
        """

        def decorator(func: Callable[Concatenate[ContextT, P], Coro[Any]]) -> GroupT:
            kwargs.setdefault('parent', self)
            result = group(name=name, cls=cls, *args, **kwargs)(func)
            self.add_command(result)
            return result

        return decorator


class Group(GroupMixin[CogT], Command[CogT, P, T]):
    """为作为子命令执行的命令实现分组协议的类。

    这个类是 :class:`.Command` 的子类，因此所有在 :class:`.Command` 中有效的选项在这里也有效。

    Attributes
    -----------
    invoke_without_command: :class:`bool`
        指示组回调是否应仅在未找到子命令时才开始解析和调用。
        用于制作一个错误处理函数以告诉用户未找到子命令或在未找到子命令的情况下具有不同的功能。
        如果这是 ``False`` ，则总是首先调用组回调。
        这意味着将执行由其参数指示的检查和解析。 默认为 ``False`` 。
    case_insensitive: :class:`bool`
        指示组的命令是否应不区分大小写。 默认为 ``False`` 。
    """

    def __init__(self, *args: Any, **attrs: Any) -> None:
        self.invoke_without_command: bool = attrs.pop('invoke_without_command', False)
        super().__init__(*args, **attrs)

    def copy(self: GroupT) -> GroupT:
        """创建此 :class:`Group` 的副本。

        Returns
        --------
        :class:`Group`
            组的新实例。
        """
        ret = super().copy()
        for cmd in self.commands:
            ret.add_command(cmd.copy())
        return ret  # type: ignore

    async def invoke(self, ctx: Context) -> None:
        ctx.invoked_subcommand = None
        ctx.subcommand_passed = None
        early_invoke = not self.invoke_without_command
        if early_invoke:
            await self.prepare(ctx)

        view = ctx.view
        previous = view.index
        view.skip_ws()
        trigger = view.get_word()

        if trigger:
            ctx.subcommand_passed = trigger
            ctx.invoked_subcommand = self.all_commands.get(trigger, None)

        if early_invoke:
            injected = hooked_wrapped_callback(self, ctx, self.callback)
            await injected(*ctx.args, **ctx.kwargs)

        ctx.invoked_parents.append(ctx.invoked_with)  # type: ignore

        if trigger and ctx.invoked_subcommand:
            ctx.invoked_with = trigger
            await ctx.invoked_subcommand.invoke(ctx)
        elif not early_invoke:
            # undo the trigger parsing
            view.index = previous
            view.previous = previous
            await super().invoke(ctx)

    async def reinvoke(self, ctx: Context, *, call_hooks: bool = False) -> None:
        ctx.invoked_subcommand = None
        early_invoke = not self.invoke_without_command
        if early_invoke:
            ctx.command = self
            await self._parse_arguments(ctx)

            if call_hooks:
                await self.call_before_hooks(ctx)

        view = ctx.view
        previous = view.index
        view.skip_ws()
        trigger = view.get_word()

        if trigger:
            ctx.subcommand_passed = trigger
            ctx.invoked_subcommand = self.all_commands.get(trigger, None)

        if early_invoke:
            try:
                await self.callback(*ctx.args, **ctx.kwargs)  # type: ignore
            except:
                ctx.command_failed = True
                raise
            finally:
                if call_hooks:
                    await self.call_after_hooks(ctx)

        ctx.invoked_parents.append(ctx.invoked_with)  # type: ignore

        if trigger and ctx.invoked_subcommand:
            ctx.invoked_with = trigger
            await ctx.invoked_subcommand.reinvoke(ctx, call_hooks=call_hooks)
        elif not early_invoke:
            # undo the trigger parsing
            view.index = previous
            view.previous = previous
            await super().reinvoke(ctx, call_hooks=call_hooks)


# Decorators

@overload
def command(
        name: str = ...,
        cls: Type[Command[CogT, P, T]] = ...,
        **attrs: Any,
) -> Callable[
    [
        Union[
            Callable[Concatenate[CogT, ContextT, P], Coro[T]],
            Callable[Concatenate[ContextT, P], Coro[T]],
        ]
    ]
    , Command[CogT, P, T]]:
    ...


@overload
def command(
        name: str = ...,
        cls: Type[CommandT] = ...,
        **attrs: Any,
) -> Callable[
    [
        Union[
            Callable[Concatenate[CogT, ContextT, P], Coro[Any]],
            Callable[Concatenate[ContextT, P], Coro[Any]],
        ]
    ]
    , CommandT]:
    ...


def command(
        name: str = MISSING,
        cls: Type[CommandT] = MISSING,
        **attrs: Any
) -> Callable[
    [
        Union[
            Callable[Concatenate[ContextT, P], Coro[Any]],
            Callable[Concatenate[CogT, ContextT, P], Coro[T]],
        ]
    ]
    , Union[Command[CogT, P, T], CommandT]]:
    """将函数转换为 :class:`.Command` 或如果使用 :func:`.group` 则转为 :class:`.Group` 的装饰器。

    默认情况下，``help`` 属性是从函数的文档字符串中自动接收的，并使用``inspect.cleandoc`` 进行清理。
    如果 docstring 是 ``bytes``，则使用 utf-8 编码将其解码为 :class:`str`。

    所有使用 :func:`.check` 添加的检查都被添加到函数中。 无法通过此装饰器提供你自己的检查。

    Parameters
    -----------
    name: :class:`str`
        用于创建命令的名称。 默认情况下，这使用未更改的函数名称。
    cls
        要构造的类。 默认情况下，这是 :class:`.Command` 。 你通常不会更改此设置。
    attrs
        传递到由 ``cls`` 表示的类的构造中的关键字参数。

    Raises
    -------
    TypeError
        如果函数不是协程或已经是命令。
    """
    if cls is MISSING:
        cls = Command  # type: ignore

    def decorator(func: Union[
        Callable[Concatenate[ContextT, P], Coro[Any]],
        Callable[Concatenate[CogT, ContextT, P], Coro[Any]],
    ]) -> CommandT:
        if isinstance(func, Command):
            raise TypeError('Callback is already a command.')
        return cls(func, name=name, **attrs)

    return decorator


@overload
def group(
        name: str = ...,
        cls: Type[Group[CogT, P, T]] = ...,
        **attrs: Any,
) -> Callable[
    [
        Union[
            Callable[Concatenate[CogT, ContextT, P], Coro[T]],
            Callable[Concatenate[ContextT, P], Coro[T]],
        ]
    ]
    , Group[CogT, P, T]]:
    ...


@overload
def group(
        name: str = ...,
        cls: Type[GroupT] = ...,
        **attrs: Any,
) -> Callable[
    [
        Union[
            Callable[Concatenate[CogT, ContextT, P], Coro[Any]],
            Callable[Concatenate[ContextT, P], Coro[Any]],
        ]
    ]
    , GroupT]:
    ...


def group(
        name: str = MISSING,
        cls: Type[GroupT] = MISSING,
        **attrs: Any,
) -> Callable[
    [
        Union[
            Callable[Concatenate[ContextT, P], Coro[Any]],
            Callable[Concatenate[CogT, ContextT, P], Coro[T]],
        ]
    ]
    , Union[Group[CogT, P, T], GroupT]]:
    """将函数转换为 :class:`.Group` 的装饰器。

     这类似于 :func:`.command` 装饰器，但 ``cls`` 参数默认设置为 :class:`Group` 。
    """
    if cls is MISSING:
        cls = Group  # type: ignore
    return command(name=name, cls=cls, **attrs)  # type: ignore


def check(predicate: Check) -> Callable[[T], T]:
    r"""向 :class:`.Command` 或其子类添加检查的装饰器。 这些检查可以通过 :attr:`.Command.checks` 访问。

    这些检查应该是接受单个参数的谓词，参数为 :class:`.Context`。
    如果检查返回类似 ``False``\ 的值，则在调用期间会引发 :exc:`.CheckFailure` 异常并将其发送到 :func:`.on_command_error` 事件。

    如果谓词中应该抛出异常，那么它应该是 :exc:`.CommandError` 的子类。
    任何不是从它子类化的异常都将被传播，而那些子类将被发送到 :func:`.on_command_error`。

    名为 ``predicate`` 的特殊属性绑定到此装饰器返回的值，以检索传递给装饰器的谓词。 这允许完成以下内省和链接：

    .. code-block:: python3

        def owner_or_permissions(**perms):
            original = commands.has_permissions(**perms).predicate
            async def extended_check(ctx):
                if ctx.guild is None:
                    return False
                return ctx.guild.owner_id == ctx.author.id or await original(ctx)
            return commands.check(extended_check)

    .. note::

        ``predicate`` 返回的函数 **始终** 是一个协程，即使原始函数不是协程。

    Examples
    ---------

    创建一个基本检查以查看命令调用者是否是你。

    .. code-block:: python3

        def check_if_it_is_me(ctx):
            return ctx.message.author.id == 114514

        @bot.command()
        @commands.check(check_if_it_is_me)
        async def only_for_me(ctx):
            await ctx.send('我知道你!')

    将常见检查转换为装饰器：

    .. code-block:: python3

        def is_me():
            def predicate(ctx):
                return ctx.message.author.id == 114514
            return commands.check(predicate)

        @bot.command()
        @is_me()
        async def only_me(ctx):
            await ctx.send('只有你!')

    Parameters
    -----------
    predicate: Callable[[:class:`Context`], :class:`bool`]
        检查是否应调用命令的谓词。
    """

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        if isinstance(func, Command):
            func.checks.append(predicate)
        else:
            if not hasattr(func, '__commands_checks__'):
                func.__commands_checks__ = []

            func.__commands_checks__.append(predicate)

        return func

    if inspect.iscoroutinefunction(predicate):
        decorator.predicate = predicate
    else:
        @functools.wraps(predicate)
        async def wrapper(ctx):
            return predicate(ctx)  # type: ignore

        decorator.predicate = wrapper

    return decorator  # type: ignore


def check_any(*checks: Check) -> Callable[[T], T]:
    r"""添加了一个 :func:`check` 来检查是否有任何通过的检查会通过，即使用逻辑 OR。

    如果所有检查都失败，则引发 :exc:`.CheckAnyFailure` 以表示失败。 它继承自 :exc:`.CheckFailure`。

    .. note::

        这个函数的 ``predicate`` 属性 **是** 一个协程。

    Parameters
    ------------
    \*checks: Callable[[:class:`Context`], :class:`bool`]
        已用 :func:`check` 装饰器装饰的检查的参数列表。

    Raises
    -------
    TypeError
        通过的检查没有用 :func:`check` 装饰器装饰。

    Examples
    ---------

    创建基本检查以查看它是机器人所有者还是频道所有者：

    .. code-block:: python3

        def is_guild_owner():
            def predicate(ctx):
                return ctx.guild is not None and ctx.guild.owner_id == ctx.author.id
            return commands.check(predicate)

        @bot.command()
        @commands.check_any(commands.is_owner(), is_guild_owner())
        async def only_for_owners(ctx):
            await ctx.send('先生你好！')
    """

    unwrapped = []
    for wrapped in checks:
        try:
            pred = wrapped.predicate
        except AttributeError:
            raise TypeError(f'{wrapped!r} 必须由 commands.check 装饰器包装') from None
        else:
            unwrapped.append(pred)

    async def predicate(ctx: Context) -> bool:
        errors = []
        for func in unwrapped:
            try:
                value = await func(ctx)
            except CheckFailure as e:
                errors.append(e)
            else:
                if value:
                    return True
        # if we're here, all checks failed
        raise CheckAnyFailure(unwrapped, errors)

    return check(predicate)


def has_role(item: Union[int, str]) -> Callable[[T], T]:
    """添加的 :func:`.check` 用于检查调用命令的成员是否具有通过指定的名称或 ID 指定的用户组。

    如果指定了字符串，则必须给出用户组的确切名称，包括大写和拼写。

    如果指定了整数，则必须提供用户组的确切 ID。

    如果消息是在私人消息上下文中调用的，则检查将返回 ``False`` 。

    此检查会引发两个特殊异常之一，如果用户缺少用户组，则为 :exc:`.MissingRole`，
    如果在私人消息中使用了 :exc:`.NoPrivateMessage`。
    两者都继承自 :exc:`.CheckFailure`。

    Parameters
    -----------
    item: Union[:class:`int`, :class:`str`]
        要检查的用户组的名称或 ID。
    """

    def predicate(ctx: Context) -> bool:
        if ctx.guild is None:
            raise NoPrivateMessage()

        # ctx.guild is None doesn't narrow ctx.author to Member
        if isinstance(item, int):
            role = qq.utils.get(ctx.author.roles, id=item)  # type: ignore
        else:
            role = qq.utils.get(ctx.author.roles, name=item)  # type: ignore
        if role is None:
            raise MissingRole(item)
        return True

    return check(predicate)


def has_any_role(*items: Union[int, str]) -> Callable[[T], T]:
    r"""A :func:`.check` 添加的内容是检查调用命令的成员是否具有指定的 **任何一个** 用户组。
    这意味着如果他们有指定了三个用户组中的一个，那么此检查将返回 `True`。

    与 :func:`.has_role`\ 类似，传入的名称或ID 必须准确无误。

    此检查会引发两个特殊异常之一，如果用户缺少所有用户组，则为 :exc:`.MissingRole`，
    如果在私人消息中使用了 :exc:`.NoPrivateMessage`。
    两者都继承自 :exc:`.CheckFailure`。

    Parameters
    -----------
    items: List[Union[:class:`str`, :class:`int`]]
        名称或 ID 的参数列表，用于检查成员是否具有用户组。

    Example
    --------

    .. code-block:: python3

        @bot.command()
        @commands.has_any_role('库开发人员', '管理员', 114514)
        async def cool(ctx):
            await ctx.send('你确实很酷')
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        # ctx.guild is None doesn't narrow ctx.author to Member
        getter = functools.partial(qq.utils.get, ctx.author.roles)  # type: ignore
        if any(getter(id=item) is not None if isinstance(item, int) else getter(name=item) is not None for item in
               items):
            return True
        raise MissingAnyRole(list(items))

    return check(predicate)


def bot_has_role(item: int) -> Callable[[T], T]:
    """类似于 :func:`.has_role`，但是是检查机器人本身是否具有用户组。

    此检查会引发两个特殊异常之一，如果机器人缺少用户组，则为 :exc:`.MissingRole`，
    如果在私人消息中使用了 :exc:`.NoPrivateMessage`。
    两者都继承自 :exc:`.CheckFailure`。
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        me = ctx.me
        if isinstance(item, int):
            role = qq.utils.get(me.roles, id=item)
        else:
            role = qq.utils.get(me.roles, name=item)
        if role is None:
            raise BotMissingRole(item)
        return True

    return check(predicate)


def bot_has_any_role(*items: int) -> Callable[[T], T]:
    """类似于 :func:`.has_any_role`，但是是检查机器人本身是否具有任何一个用户组。

    此检查会引发两个特殊异常之一，如果机器人缺少所有用户组，则为 :exc:`.MissingRole`，
    如果在私人消息中使用了 :exc:`.NoPrivateMessage`。
    两者都继承自 :exc:`.CheckFailure`。
    """

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()

        me = ctx.me
        getter = functools.partial(qq.utils.get, me.roles)
        if any(getter(id=item) is not None if isinstance(item, int) else getter(name=item) is not None for item in
               items):
            return True
        raise BotMissingAnyRole(list(items))

    return check(predicate)


def dm_only() -> Callable[[T], T]:
    """一个 :func:`.check`
    表示这个命令只能在 DM 上下文中使用。 使用该命令时只允许私信。

    这个检查引发一个特殊的异常， :exc:`.PrivateMessageOnly`，它继承自 :exc:`.CheckFailure`。
    """

    def predicate(ctx: Context) -> bool:
        if ctx.guild is not None:
            raise PrivateMessageOnly()
        return True

    return check(predicate)


def guild_only() -> Callable[[T], T]:
    """一个 :func:`.check` 表示这个命令只能在频道上下文中使用。 基本上，使用该命令时不允许私人消息。

    这个检查引发一个特殊的异常，:exc:`.NoPrivateMessage`，它是从 :exc:`.CheckFailure` 继承而来的。
    """

    def predicate(ctx: Context) -> bool:
        if ctx.guild is None:
            raise NoPrivateMessage()
        return True

    return check(predicate)


def is_owner() -> Callable[[T], T]:
    """一个 :func:`.check` 检查调用此命令的人是否是机器人的所有者。

    这是由 :meth:`.Bot.is_owner` 提供支持的。

    这个检查引发了一个特殊的异常， :exc:`.NotOwner`，它派生自 :exc:`.CheckFailure`。
    """

    async def predicate(ctx: Context) -> bool:
        if not await ctx.bot.is_owner(ctx.author):
            raise NotOwner('你不拥有此机器人。')
        return True

    return check(predicate)


def cooldown(rate: int, per: float, type: Union[BucketType, Callable[[Message], Any]] = BucketType.default) -> Callable[
    [T], T]:
    """为 :class:`.Command` 添加冷却时间的装饰器

    冷却时间允许命令在特定时间范围内仅使用特定次数。
    这些冷却时间可以基于每个公会、每个频道、每个用户、
    每个用户组或全局基础。由第三个参数 ``type`` 表示，它必须是枚举类型 :class:`.BucketType`。

    如果冷却被触发，则 :exc:`.CommandOnCooldown` 在 :func:`.on_command_error` 和本地错误处理程序中被触发。

    一个命令只能有一个冷却时间。

    Parameters
    ------------
    rate: :class:`int`
        命令在触发冷却之前可以使用的次数。
    per: :class:`float`
        触发时等待冷却的秒数。
    type: Union[:class:`.BucketType`, Callable[[:class:`.Message`], Any]]
        冷却时间的类型。如果可调用，则应返回映射的键。
    """

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        if isinstance(func, Command):
            func._buckets = CooldownMapping(Cooldown(rate, per), type)
        else:
            func.__commands_cooldown__ = CooldownMapping(Cooldown(rate, per), type)
        return func

    return decorator  # type: ignore


def dynamic_cooldown(cooldown: Union[BucketType, Callable[[Message], Any]], type: BucketType = BucketType.default) -> \
        Callable[[T], T]:
    """为 :class:`.Command` 添加动态冷却时间的装饰器

    这与 :func:`.cooldown` 的不同之处在于
    它接受一个 :class:`.qq.Message` 类型的单个参数并且必须返回一个 :class:`.Cooldown`
    或 ``None`` 的函数。如果返回 ``None`` ，则该冷却时间被有效绕过。

    冷却时间允许命令在特定时间范围内仅使用特定次数。
    这些冷却时间可以基于每个公会、每个频道、每个用户、每个角色或全局基础。
    由 ``type`` 的第三个参数表示，它必须是枚举类型 :class:`.BucketType` 。

    如果冷却被触发，则 :exc:`.CommandOnCooldown` 在 :func:`.on_command_error` 和本地错误处理程序中被触发。

    一个命令只能有一个冷却时间。

    Parameters
    ------------
    cooldown: Callable[[:class:`.qq.Message`], Optional[:class:`.Cooldown`]]
        一个接收消息并返回冷却时间的函数，该冷却时间将应用于此调用，如果应该绕过冷却时间，则返回 ``None`` 。
    type: :class:`.BucketType`
        冷却时间的类型。
    """
    if not callable(cooldown):
        raise TypeError("必须提供可调用")

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        if isinstance(func, Command):
            func._buckets = DynamicCooldownMapping(cooldown, type)
        else:
            func.__commands_cooldown__ = DynamicCooldownMapping(cooldown, type)
        return func

    return decorator  # type: ignore


def max_concurrency(number: int, per: BucketType = BucketType.default, *, wait: bool = False) -> Callable[[T], T]:
    """为 :class:`.Command` 或其子类添加最大并发的装饰器。

    这使你可以同时只允许一定数量的命令调用，例如，如果命令耗时过长或一次只有一个用户可以使用它。
    这与冷却时间不同，因为没有设定的等待期或令牌桶——只有设定数量的人可以运行命令。

    Parameters
    -------------
    number: :class:`int`
        可以同时运行的此命令的最大调用次数。
    per: :class:`.BucketType`
        此并发所基于的存储桶，例如``BucketType.guild`` 将允许每个公会最多使用 ``number`` 次。
    wait: :class:`bool`
        命令是否应该等待队列结束。
        如果这被设置为 ``False`` 则不是等到命令可以再次运行，该命令会引发 :exc:`.MaxConcurrencyReached` 到其错误处理程序。
        如果这被设置为 ``True`` 则命令会等待直到它可以被执行。
    """

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        value = MaxConcurrency(number, per=per, wait=wait)
        if isinstance(func, Command):
            func._max_concurrency = value
        else:
            func.__commands_max_concurrency__ = value
        return func

    return decorator  # type: ignore


def before_invoke(coro) -> Callable[[T], T]:
    """将协程注册为调用前钩的装饰器。

    这允许你在调用钩子之前引用一个不必在同一个齿轮中的几个命令。

    Example
    ---------

    .. code-block:: python3

        async def record_usage(ctx):
            print(ctx.author, 'used', ctx.command)

        @bot.command()
        @commands.before_invoke(record_usage)
        async def who(ctx):
            await ctx.send('我是机器人')

        class What(commands.Cog):

            @commands.command()
            async def where(self, ctx): # Output: <Nothing>
                await ctx.send('在QQ上')

            @commands.command()
            async def why(self, ctx): # Output: <Nothing>
                await ctx.send('因为有人做了我出来')

        bot.add_cog(What())
    """

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        if isinstance(func, Command):
            func.before_invoke(coro)
        else:
            func.__before_invoke__ = coro
        return func

    return decorator  # type: ignore


def after_invoke(coro) -> Callable[[T], T]:
    """将协程注册为调用后钩的装饰器。

    这允许你在调用钩子后引用多个命令，这些命令不必位于同一个 cog 中。
    """

    def decorator(func: Union[Command, CoroFunc]) -> Union[Command, CoroFunc]:
        if isinstance(func, Command):
            func.after_invoke(coro)
        else:
            func.__after_invoke__ = coro
        return func

    return decorator  # type: ignore



from __future__ import annotations

from typing import Any, Dict, Optional, TYPE_CHECKING, Type, TypeVar, Union
import re

from .asset import AssetMixin
from .error import InvalidArgument

__all__ = (
    'PartialEmoji',
)

if TYPE_CHECKING:
    from .state import ConnectionState
    from datetime import datetime
    from .types.message import PartialEmoji as PartialEmojiPayload


class _EmojiTag:
    __slots__ = ()

    id: int

    def _to_partial(self) -> PartialEmoji:
        raise NotImplementedError


PE = TypeVar('PE', bound='PartialEmoji')


class PartialEmoji(_EmojiTag, AssetMixin):
    """代表 “部分” 表情符号。
    该模型将在两种情况下给出:

    - “原始”数据事件，例如 :func:`on_raw_reaction_add`
    - 机器人无法看到的自定义表情符号，例如 :attr:`Message.reactions`

    .. container:: operations

        .. describe:: x == y

            检查两个表情符号是否相同。

        .. describe:: x != y

            检查两个表情符号是否不同。

        .. describe:: hash(x)

            返回表情符号的哈希值。

        .. describe:: str(x)

            返回为 QQ 渲染的表情符号。


    Attributes
    -----------
    custom: :class:`bool`
        表情是否是 QQ 自定义表情。
    id: :class:`str`
        自定义表情符号的 ID（如果适用）。
    """

    __slots__ = ('animated', 'name', 'id', '_state', 'custom')

    _CUSTOM_EMOJI_RE = re.compile(r'<?(?P<animated>a)?:?(?P<name>[A-Za-z0-9\_]+):(?P<id>[0-9]{13,20})>?')

    if TYPE_CHECKING:
        id: Optional[str]

    def __init__(self, *, custom: bool, id: str = None):
        self.custom = custom
        self.animated = False
        self.name = 'emoji'
        self.id = id
        self._state: Optional[ConnectionState] = None

    @classmethod
    def from_dict(cls: Type[PE], data: Union[PartialEmojiPayload, Dict[str, Any]]) -> PE:
        return cls(
            id=data.get('id'),
            custom=True if data.get('type') == 1 else 0,
        )

    @classmethod
    def from_str(cls: Type[PE], value: str) -> PE:
        """将表情符号的 QQ 字符串表示形式转换为 :class:`PartialEmoji`。
        接受的格式是：

        - ``a:name:id``
        - ``<a:name:id>``
        - ``name:id``
        - ``<:name:id>``

        如果格式不匹配，则假定它是一个 unicode 表情符号。

        Parameters
        ------------
        value: :class:`str`
            表情符号的字符串表示。

        Returns
        --------
        :class:`PartialEmoji`
            此字符串中的表情符号。
        """
        match = cls._CUSTOM_EMOJI_RE.match(value)
        if match is not None:
            groups = match.groupdict()
            emoji_id = groups['id']
            return cls(id=emoji_id, custom=True)

        return cls(id=value, custom=False)

    def to_dict(self) -> Dict[str, Any]:
        o: Dict[str, Any] = {'id': self.id, 'type': '1' if self.custom else '2'}
        return o

    def _to_partial(self) -> PartialEmoji:
        return self

    @classmethod
    def with_state(
            cls: Type[PE], state: ConnectionState, *, custom: bool, id: str = None
    ) -> PE:
        self = cls(custom=custom, id=id)
        self._state = state
        return self

    def __str__(self) -> str:
        if self.id is None:
            return self.name
        if self.animated:
            return f'<a:{self.name}:{self.id}>'
        return f'<{self.name}:{self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} id={self.id} type={"1" if self.custom else "2"}>'

    def __eq__(self, other: Any) -> bool:
        if self.is_unicode_emoji():
            return isinstance(other, PartialEmoji) and self.name == other.name

        if isinstance(other, _EmojiTag):
            return self.id == other.id
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.id, self.name))

    def is_custom_emoji(self) -> bool:
        """:class:`bool`: 检查这是否是自定义的非 Unicode 表情符号。"""
        return self.custom

    def is_unicode_emoji(self) -> bool:
        """:class:`bool`: 检查这是否是 Unicode 表情符号。"""
        return not self.custom

    def _as_reaction(self) -> str:
        if self.id is None:
            return self.name
        return f'{self.name}:{self.id}'

    async def read(self) -> bytes:
        if self.is_unicode_emoji():
            raise InvalidArgument('PartialEmoji 不是自定义表情符号')

        return await super().read()



from __future__ import annotations

import datetime
from typing import (
    Dict,
    List,
    Union,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Any,
    Sequence,
    overload, Literal,
)

from . import utils, abc
from .role import Role
from .member import Member
from .error import InvalidData
from .colour import Colour
from .error import InvalidArgument, ClientException
from .channel import *
from .channel import _guild_channel_factory, _channel_factory
from .mixins import Hashable
from .user import User
from .iterators import MemberIterator
from .asset import Asset
from .file import File

MISSING = utils.MISSING

__all__ = (
    'Guild',
)

if TYPE_CHECKING:
    from .enum import ChannelType
    from .state import ConnectionState
    from .types.guild import Guild as GuildPayload
    from .channel import TextChannel, CategoryChannel, AppChannel, LiveChannel, ThreadChannel
    from .state import ConnectionState

    GuildChannel = Union[VoiceChannel, TextChannel, CategoryChannel, AppChannel, LiveChannel, ThreadChannel]
    VocalGuildChannel = Union[VoiceChannel]
    ByCategoryItem = Tuple[Optional[CategoryChannel], List[GuildChannel]]


class Guild(Hashable):
    """代表一个 QQ guild.
    这在官方 QQ UI 中称为  ``频道``  。

    .. container:: operations

        .. describe:: x == y
            检查两个 guild 是否相等。
        .. describe:: x != y
            检查两个 guild 是否不相等。
        .. describe:: hash(x)
            返回 guild 的哈希值。
        .. describe:: str(x)
            返回 guild 的名称。

    Attributes
    ----------
    name: :class:`str`
        guild 名称。
    id: :class:`int`
        guild 的 ID。
    owner_id: :class:`int`
        guild 所有者的ID。使用 :attr:`Guild.owner` 代替。
    unavailable: :class:`bool`
        指示频道是否不可用。如果这是 ``True`` 则 :attr:`Guild.id` 之外的其他属性的可靠性很小，它们可能都是 ``None``。
        如果频道不可用，最好不要对频道做任何事情。
    max_members: Optional[:class:`int`]
        guild 成员的最大数量。

        .. note::

            该属性只能通过 :meth:`.Client.fetch_guild` 获得。

    description: Optional[:class:`str`]
        guild 的说明。
    """

    __slots__ = (
        'id',
        'name',
        'icon',
        'owner_id',
        '_member_count',
        'max_members',
        'description',
        'joined_at',
        '_channels',
        '_members',
        '_roles',
        '_state',
        '_large',
        'unavailable'
    )

    def __init__(self, data: GuildPayload, state: ConnectionState):
        self._channels: Dict[int, GuildChannel] = {}
        self._members: Dict[int, Member] = {}
        self._state: ConnectionState = state
        self._from_data(data)

    def _add_role(self, role: Role, /) -> None:
        self._roles[role.id] = role

    def _remove_role(self, role_id: int, /) -> Role:
        # this raises KeyError if it fails.
        role = self._roles.pop(role_id)
        return role

    def _from_data(self, guild: GuildPayload) -> None:
        self.id = int(guild.get('id'))
        self.name = guild.get('name')
        self.icon = guild.get('icon')
        self.owner_id = guild.get('owner_id')
        if self.owner_id:
            self.owner_id = int(self.owner_id)
        self._member_count = guild.get('member_count')
        self.max_members = guild.get('max_members')
        self.description = guild.get('description')
        self.joined_at = guild.get('joined_at')
        self.unavailable: bool = guild.get('unavailable', False)
        self._roles: Dict[int, Role] = {}
        state = self._state  # speed up attribute access
        self._large: Optional[bool] = None if self._member_count is None else self._member_count >= 250
        for r in guild.get('roles', []):
            role = Role(guild=self, data=r, state=state)
            self._roles[role.id] = role
        self._sync()

    def _add_channel(self, channel: GuildChannel, /) -> None:
        self._channels[channel.id] = channel

    def _remove_channel(self, channel: GuildChannel, /) -> None:
        self._channels.pop(channel.id, None)

    def __str__(self) -> str:
        return self.name or ''

    def __repr__(self) -> str:
        attrs = (
            ('id', self.id),
            ('name', self.name),
            ('member_count', getattr(self, '_member_count', None)),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'<Guild {inner}>'

    def _sync(self) -> None:
        # I know it's jank to put a sync requests here,
        # but QQ just does not give all the info about guilds unless you requests it
        channels, roles, members = self._state.http.sync_guild_channels_roles(self.id)
        if members:
            for mdata in members:
                member = Member(data=mdata, guild=self, state=self._state)
                self._add_member(member)
        else:
            member = Member(data=self._state.http.sync_get_bot_member(self.id, self._state.user.id),
                            guild=self, state=self._state)
            self._add_member(member)
        if 'roles' in roles:
            for r in roles['roles']:
                role = Role(guild=self, data=r, state=self._state)
                self._roles[role.id] = role
        for c in channels:
            factory, ch_type = _guild_channel_factory(c['type'])
            if factory:
                self._add_channel(factory(guild=self, data=c, state=self._state))  # type: ignore

    @property
    def channels(self) -> List[GuildChannel]:
        """List[:class:`abc.GuildChannel`]: 属于该频道的子频道列表。"""
        return list(self._channels.values())

    @property
    def shard_id(self) -> int:
        """:class:`int`: 如果适用，返回此频道的分片 ID。"""
        count = self._state.shard_count
        if count is None:
            return 0
        return (self.id >> 22) % count

    @property
    def owner(self) -> Optional[Member]:
        """Optional[:class:`Member`]: 拥有频道的成员。"""
        return self.get_member(self.owner_id)

    @property
    def members(self) -> List[Member]:
        """List[:class:`Member`]: 属于该频道的成员列表。"""
        return list(self._members.values())

    def get_role(self, role_id: int, /) -> Optional[Role]:
        """返回具有给定 ID 的 Role。

        Parameters
        -----------
        role_id: :class:`int`
            要搜索的 ID。

        Returns
        --------
        Optional[:class:`Role`]
            Role 或如果未找到，则  ``None`` 。
        """
        return self._roles.get(role_id)

    # @property
    # def large(self) -> bool:
    #     """:class:`bool`: 指示频道是否是 ``大型`` 频道。
    #
    #     一个大型频道被定义为拥有超过 ``large_threshold`` 计数成员，对于这个库，它被设置为 250。
    #     """
    #     if self._large is None:
    #         try:
    #             return self._member_count >= 250
    #         except AttributeError:
    #             return len(self._members) >= 250
    #     return self._large

    @property
    def me(self) -> Member:
        """:class:`Member`: 类似于 :attr:`Client.user` ，除了它是 :class:`Member` 的一个实例。
        这主要用于获取您自己的 Member 版本。
        """
        self_id = self._state.user.id
        # The self member is *always* cached
        return self.get_member(self_id)  # type: ignore

    @property
    def text_channels(self) -> List[TextChannel]:
        """List[:class:`TextChannel`]: 属于该频道的文本频道列表。这是按位置排序的，从上到下按 UI 顺序排列。
        """
        r = [ch for ch in self._channels.values() if isinstance(ch, TextChannel)]
        r.sort(key=lambda c: (c.position, c.id))
        return r

    @property
    def categories(self) -> List[CategoryChannel]:
        """List[:class:`CategoryChannel`]: 属于该频道的类别列表。这是按位置排序的，从上到下按 UI 顺序排列。
        """
        r = [ch for ch in self._channels.values() if isinstance(ch, CategoryChannel)]
        r.sort(key=lambda c: (c.position, c.id))
        return r

    def by_category(self) -> List[ByCategoryItem]:
        """返回每个 :class:`CategoryChannel` 及其关联的频道。
        这些频道和类别按官方 QQ UI 顺序排序。
        如果频道没有类别，则元组的第一个元素是  ``None`` 。

        Returns
        --------
        List[Tuple[Optional[:class:`CategoryChannel`], List[:class:`abc.GuildChannel`]]]:
            类别及其关联的频道。
        """
        grouped: Dict[Optional[int], List[GuildChannel]] = {}
        for channel in self._channels.values():
            if isinstance(channel, CategoryChannel):
                grouped.setdefault(channel.id, [])
                continue

            try:
                grouped[channel.category_id].append(channel)
            except KeyError:
                grouped[channel.category_id] = [channel]

        def key(t: ByCategoryItem) -> Tuple[Tuple[int, int], List[GuildChannel]]:
            k, v = t
            return (k.position, k.id) if k else (-1, -1), v

        _get = self._channels.get
        as_list: List[ByCategoryItem] = [(_get(k), v) for k, v in grouped.items()]  # type: ignore
        as_list.sort(key=key)
        for _, channels in as_list:
            channels.sort(key=lambda c: (c._sorting_bucket, c.position, c.id))
        return as_list

    def _resolve_channel(self, id: Optional[int], /) -> Optional[Union[GuildChannel,]]:
        if id is None:
            return

        return self._channels.get(id)

    def get_channel(self, channel_id: int, /) -> Optional[GuildChannel]:
        """返回具有给定 ID 的频道。

        Parameters
        -----------
        channel_id: :class:`int`
            要搜索的 ID。

        Returns
        --------
        Optional[:class:`.abc.GuildChannel`]
            返回的频道或 ``None``（如果未找到）。
        """
        return self._channels.get(channel_id)

    def get_member(self, user_id: int, /) -> Optional[Member]:
        """返回具有给定 ID 的成员。

        Parameters
        -----------
        user_id: :class:`int`
            要搜索的 ID。

        Returns
        --------
        Optional[:class:`Member`]
            返回成员或如果未找到  ``None``  。
        """
        return self._members.get(user_id)

    @property
    def roles(self) -> List[Role]:
        """List[:class:`Role`]: 以层级顺序返回频道身份组的 :class:`list`。此列表的第一个元素将是层次结构中的最低身份组。
        """
        return sorted(self._roles.values())

    def _add_member(self, member: Member, /) -> None:
        self._members[member.id] = member

    def _remove_member(self, member: Member, /) -> None:
        self._members.pop(member.id, None)

    @property
    def chunked(self) -> bool:
        count = getattr(self, '_member_count', None)
        if count is None:
            return False
        return count == len(self._members)

    def get_member_named(self, name: str, /) -> Optional[Member]:
        """返回找到的第一个与提供的名称匹配的成员。
        如果传递了昵称，则通过昵称查找它。
        如果没有找到成员，则返回  ``None`` 。

        Parameters
        -----------
        name: :class:`str`
            要查找的成员的名称。

        Returns
        --------
        Optional[:class:`Member`]
            此频道中具有关联名称的成员。如果未找到，则返回  ``None`` 。
        """
        members = self.members
        result = utils.get(members, name=name[:-5])
        if result is not None:
            return result

        def pred(m: Member) -> bool:
            return m.nick == name or m.name == name

        return utils.find(pred, members)

    @property
    def member_count(self) -> int:
        """:class:`int`: 无论是否完全加载，都返回真实的成员计数。

        .. warning::

            由于 QQ 的限制，为了使该属性保持最新和准确，它需要指定 :attr:`Intents.members` 。
        """
        return self._member_count

    async def create_text_channel(
            self,
            name: str,
            *,
            reason: Optional[str] = None,
            category: Optional[CategoryChannel] = None,
            position: int = MISSING,
    ) -> TextChannel:
        """|coro|
        为频道创建一个 :class:`TextChannel` 。

        .. note::

            创建指定位置的频道不会更新其他频道的位置以效仿。需要对 :meth:`~TextChannel.edit` 进行后续调用以更新频道在频道列表中的位置。

        Examples
        ----------
        创建基本频道：

        .. code-block:: python3

            channel = await guild.create_text_channel('cool-channel')

        Parameters
        -----------
        name: :class:`str`
            频道的名称。
        category: Optional[:class:`CategoryChannel`]
            将新创建的子频道置于其下的类别。
        position: :class:`int`
            在子频道列表中的位置。这是一个从 0 开始的数字。例如顶部子频道是位置 0。
        reason: Optional[:class:`str`]
            创建此频道的原因。

        Raises
        -------
        Forbidden
            您没有创建此频道的适当权限。
        HTTPException
            创建频道失败。
        Returns
        -------
        :class:`TextChannel`
            刚刚创建的频道。
        """

        options = {}
        if position is not MISSING:
            options['position'] = position

        data = await self._create_channel(
            name, channel_type=ChannelType.text, category=category, reason=reason, **options
        )
        channel = TextChannel(state=self._state, guild=self, data=data)

        # temporarily add to the cache
        self._channels[channel.id] = channel
        return channel

    def _create_channel(
            self,
            name: str,
            channel_type: ChannelType,
            category: Optional[CategoryChannel] = None,
            **options: Any,
    ):
        parent_id = category.id if category else None
        return self._state.http.create_channel(
            self.id, channel_type.value, name=name, parent_id=parent_id, **options
        )

    async def create_live_channel(
            self,
            name: str,
            *,
            position: int = MISSING,
            category: Optional[CategoryChannel] = None,
            reason: Optional[str] = None,
    ) -> LiveChannel:
        """|coro|
        这类似于 :meth:`create_live_channel` ，除了创建一个 :class:`LiveChannel` 。

        Parameters
        -----------
        name: :class:`str`
            频道的名称。
        category: Optional[:class:`LiveChannel`]
            将新创建的频道置于其下的类别。
        position: :class:`int`
            在频道列表中的位置。这是一个从 0 开始的数字。例如顶部通道是位置 0。
        reason: Optional[:class:`str`]
            创建此频道的原因。

        Raises
        ------
        Forbidden
            您没有创建此频道的适当权限。
        HTTPException
            创建频道失败。

        Returns
        -------
        :class:`LiveChannel`
            刚刚创建的频道。
        """

        options: Dict[str, Any] = {}
        if position is not MISSING:
            options['position'] = position

        data = await self._create_channel(
            name, channel_type=ChannelType.live, category=category, reason=reason, **options
        )
        channel = LiveChannel(state=self._state, guild=self, data=data)

        # temporarily add to the cache
        self._channels[channel.id] = channel
        return channel

    async def create_app_channel(
            self,
            name: str,
            *,
            position: int = MISSING,
            category: Optional[CategoryChannel] = None,
            reason: Optional[str] = None,
    ) -> AppChannel:
        """|coro|
        这类似于 :meth:`create_app_channel` ，除了生成一个 :class:`AppChannel`。

        Parameters
        -----------
        name: :class:`str`
            频道的名称。
        category: Optional[:class:`AppChannel`]
            将新创建的频道置于其下的类别。
        position: :class:`int`
            在频道列表中的位置。这是一个从 0 开始的数字。例如顶部通道是位置 0。
        reason: Optional[:class:`str`]
            创建此频道的原因。

        Raises
        ------
        Forbidden
            您没有创建此频道的适当权限。
        HTTPException
            创建频道失败。

        Returns
        -------
        :class:`AppChannel`
            刚刚创建的频道。
        """

        options: Dict[str, Any] = {}
        if position is not MISSING:
            options['position'] = position

        data = await self._create_channel(
            name, channel_type=ChannelType.live, category=category, reason=reason, **options
        )
        channel = AppChannel(state=self._state, guild=self, data=data)

        # temporarily add to the cache
        self._channels[channel.id] = channel
        return channel

    async def create_thread_channel(
            self,
            name: str,
            *,
            position: int = MISSING,
            category: Optional[CategoryChannel] = None,
            reason: Optional[str] = None,
    ) -> ThreadChannel:
        """|coro|
        这类似于 :meth:`create_thread_channel` ，除了生成一个 :class:`ThreadChannel`。

        Parameters
        -----------
        name: :class:`str`
            频道的名称。
        category: Optional[:class:`ThreadChannel`]
            将新创建的频道置于其下的类别。
        position: :class:`int`
            在频道列表中的位置。这是一个从 0 开始的数字。例如顶部通道是位置 0。
        reason: Optional[:class:`str`]
            创建此频道的原因。

        Raises
        ------
        Forbidden
            您没有创建此频道的适当权限。
        HTTPException
            创建频道失败。

        Returns
        -------
        :class:`LiveChannel`
            刚刚创建的频道。
        """

        options: Dict[str, Any] = {}
        if position is not MISSING:
            options['position'] = position

        data = await self._create_channel(
            name, channel_type=ChannelType.live, category=category, reason=reason, **options
        )
        channel = ThreadChannel(state=self._state, guild=self, data=data)

        # temporarily add to the cache
        self._channels[channel.id] = channel
        return channel

    async def create_category(
            self,
            name: str,
            *,
            reason: Optional[str] = None,
            position: int = MISSING,
    ) -> CategoryChannel:
        """|coro|
        与 :meth:`create_text_channel` 相同，除了创建一个 :class:`CategoryChannel`。

        .. note::
            此函数不支持 ``category`` 参数，因为类别不能有类别。

        Raises
        ------
        Forbidden
            您没有创建此频道的适当权限。
        HTTPException
            创建频道失败。

        Returns
        -------
        :class:`CategoryChannel`
            刚刚创建的频道。
        """
        options: Dict[str, Any] = {}
        if position is not MISSING:
            options['position'] = position

        data = await self._create_channel(
            name, channel_type=ChannelType.category, reason=reason, **options
        )
        channel = CategoryChannel(state=self._state, guild=self, data=data)

        # temporarily add to the cache
        self._channels[channel.id] = channel
        return channel

    create_category_channel = create_category

    async def fetch_channels(self) -> Sequence[GuildChannel]:
        """|coro|
        检索频道拥有的所有 :class:`abc.GuildChannel`。

        .. note::
            该方法是一个 API 调用。 对于一般用途，请考虑 :attr:`channels`。

        Raises
        -------
        InvalidData
            从 QQ 接收到未知的频道类型。
        HTTPException
            检索频道失败。

        Returns
        -------
        Sequence[:class:`abc.GuildChannel`]
            频道内的所有频道。
        """
        data = await self._state.http.get_all_guild_channels(self.id)

        def convert(d):
            factory, ch_type = _guild_channel_factory(d['type'])
            if factory is None:
                raise InvalidData('Unknown channel type {type} for channel ID {id}.'.format_map(d))

            channel = factory(guild=self, state=self._state, data=d)
            return channel

        return [convert(d) for d in data]

    def fetch_members(self, *, limit: int = 1000) -> MemberIterator:
        """检索一个 :class:`.AsyncIterator` 来接收频道的成员。为了使用它，必须启用:meth:`Intents.members`。

        .. note::
            该方法是一个 API 调用。对于一般用法，请考虑 :attr:`members`。

        Parameters
        ----------
        limit: Optional[:class:`int`]
            要检索的成员数。默认为 1000。传递 ``无`` 以获取所有成员。请注意，这可能很慢。

        Raises
        ------
        ClientException
            成员意图未启用。
        HTTPException
            获取成员失败。

        Yields
        ------
        :class:`.Member`
            已解析成员数据的成员。

        Examples
        --------
        用法  ::

            async for member in guild.fetch_members(limit=150):
                print(member.name)

        展平成一个列表 ::

            members = await guild.fetch_members(limit=150).flatten()
            # 成员现在是一个Member列表 ...

        """

        if not self._state._intents.members:
            raise ClientException('Intents.members must be enabled to use this.')

        return MemberIterator(self, limit=limit)

    async def fetch_member(self, member_id: int, /) -> Member:
        """|coro|
        从频道 ID 和成员 ID 中检索:class:`Member`。

        .. note::
            该方法是一个 API 调用。如果您启用了 :attr:`Intents.members` 和成员缓存，请考虑使用 :meth:`get_member`。

        Parameters
        -----------
        member_id: :class:`int`
            要从中获取的成员 ID。

        Raises
        -------
        Forbidden
            您无权访问频道。
        HTTPException
            获取成员失败。

        Returns
        --------
        :class:`Member`
            来自会员 ID 的会员。
        """
        data = await self._state.http.get_member(self.id, member_id)
        return Member(data=data, state=self._state, guild=self)

    async def fetch_channel(self, channel_id: int, /) -> GuildChannel:
        """|coro|
        检索具有指定 ID 的 :class:`.abc.GuildChannel`。

        .. note::

            该方法是一个 API 调用。对于一般用法，请考虑 :meth:`get_channel`。

        Raises
        -------
        :exc:`.InvalidData`
            从 QQ 接收到未知的频道类型或频道所属的频道与此对象中指向的频道不同。
        :exc:`.HTTPException`
            检索频道失败。
        :exc:`.NotFound`
            无效的频道 ID。
        :exc:`.Forbidden`
            您无权获取此频道。

        Returns
        --------
        :class:`.abc.GuildChannel`
            来自 ID 的频道。
        """
        data = await self._state.http.get_channel(channel_id)

        factory, ch_type = _channel_factory(data['type'])
        if factory is None:
            raise InvalidData('Unknown channel type {type} for channel ID {id}.'.format_map(data))

        guild_id = int(data['guild_id'])
        if self.id != guild_id:
            raise InvalidData('Guild ID resolved to a different guild')

        channel: GuildChannel = factory(guild=self, state=self._state, data=data)  # type: ignore
        return channel

    async def fetch_roles(self) -> List[Role]:
        """|coro|
        检索频道拥有的所有 :class:`Role`。
        .. note::

            该方法是一个 API 调用。对于一般用法，请考虑 :attr:`roles`。
        .. versionadded:: 1.3

        Raises
        -------
        HTTPException
            检索身份组失败。

        Returns
        -------
        List[:class:`Role`]
            频道中的所有身份组。
        """
        data = await self._state.http.get_roles(self.id)
        return [Role(guild=self, state=self._state, data=d) for d in data]

    @overload
    async def create_role(
            self,
            *,
            reason: Optional[str] = ...,
            name: str = ...,
            colour: Union[Colour, int] = ...,
            hoist: bool = ...,
            mentionable: bool = ...,
    ) -> Role:
        ...

    @overload
    async def create_role(
            self,
            *,
            reason: Optional[str] = ...,
            name: str = ...,
            color: Union[Colour, int] = ...,
            hoist: bool = ...,
            mentionable: bool = ...,
    ) -> Role:
        ...

    async def create_role(
            self,
            *,
            name: str = MISSING,
            color: Union[Colour, int] = MISSING,
            colour: Union[Colour, int] = MISSING,
            hoist: bool = MISSING,
            mentionable: bool = MISSING,
            reason: Optional[str] = None,
    ) -> Role:
        """|coro|
        为频道创建一个身份组。
        
        Parameters
        -----------
        name: :class:`str`
            身份组名称。
        colour: Union[:class:`Colour`, :class:`int`]
            身份组的颜色。默认为 :meth:`Colour.default` 。这也是 ``Color`` 的别名。
        hoist: :class:`bool`
            指示身份组是否应单独显示在成员列表中。默认为 ``False``。
        mentionable: :class:`bool`
            指示身份组是否应该被其他人提及。默认为 ``False``。
        reason: Optional[:class:`str`]
            创建此身份组的原因。
            
        Raises
        -------
        Forbidden
            您无权创建该身份组。
        HTTPException
            创建身份组失败。
        InvalidArgument
            给出了无效的关键字参数。
            
        Returns
        --------
        :class:`Role`
            新创建的身份组。
        """
        fields: Dict[str, Any] = {}

        actual_colour = colour or color or Colour.default()
        if isinstance(actual_colour, int):
            fields['color'] = actual_colour
        else:
            fields['color'] = actual_colour.value

        if hoist is not MISSING:
            fields['hoist'] = hoist

        if mentionable is not MISSING:
            fields['mentionable'] = mentionable

        if name is not MISSING:
            fields['name'] = name

        data = await self._state.http.create_role(self.id, reason=reason, **fields)
        role = Role(guild=self, data=data, state=self._state)

        return role

    async def kick(self, user: Member, *, reason: Optional[str] = None) -> None:
        """|coro|
        将一个用户踢出公会。

        Parameters
        -----------
        user: :class:`Member`
            踢出的用户。
        reason: Optional[:class:`str`]
            用户被踢的原因。

        Raises
        -------
        Forbidden
            您没有正确的踢出权限。
        HTTPException
            踢出失败。
        """
        await self._state.http.kick(user.id, self.id, reason=reason)

    async def mute_member(
            self,
            user: Member,
            *,
            duration: Union[datetime.datetime, int] = 10,
            reason: Optional[str] = None,
    ) -> None:
        """|coro|
        频道指定成员禁言。
        需要使用的token对应的用户(或机器人)具备管理员权限。

        Parameters
        -----------
        user: :class:`qq.User`
            这个频道禁言的用户。
        duration: Union[:class:`datetime.datetime`, :class:`int`]
            禁言的时间，可以是结束时间的一个 :class:`datetime.datetime` ， 也可以是持续的秒数。
        reason: Optional[:class:`str`]
            禁言的原因。

        Raises
        -------
        Forbidden
            你没有适当的权限。
        HTTPException
            禁言失败。
        """
        await self._state.http.mute_member(user.id, self.id, duration, reason=reason)

    async def mute_guild(
            self,
            *,
            duration: Union[datetime.datetime, int] = 10,
            reason: Optional[str] = None,
    ) -> None:
        """|coro|
        频道全局禁言。
        需要使用的token对应的用户(或机器人)具备管理员权限。

        Parameters
        -----------
        duration: Union[:class:`datetime.datetime`, :class:`int`]
            禁言的时间，可以是结束时间的一个 :class:`datetime.datetime` ， 也可以是持续的秒数。
        reason: Optional[:class:`str`]
            禁言的原因。

        Raises
        -------
        Forbidden
            你没有适当的权限。
        HTTPException
            禁言失败。
        """
        await self._state.http.mute_guild(self.id, duration, reason=reason)
